#!/usr/local/bin/python
'World-Aggregator'

# Core.
from .dblib import flagNames, INDEX_FILE, LocalIndex, Zonefix

from .zonelib import ZoneReader
from .roomlib import RoomReader
from .objlib import ObjectReader
from .moblib import MobileReader

from .dblib import directionName, directionIndicator, directionByName
from .dblib import doorState, wearPosition
from .dblib import zoneFlagNames, resetModeName
from .dblib import roomFlagNames, roomSectorName

class WorldBase:
    class Mixin:
        # For access to proper direct manipulation of the world instance.
        def lock(self):
            pass
        def unlock(self):
            pass

        def __getitem__(self, *args):
            return self.zones.__getitem__(*args)
        def __setitem__(self, *args):
            return self.zones.__setitem__(*args)

        def __repr__(self):
            return '<World %r (%d zones)>' % (self.fileBase, len(self.zones))

        def __iter__(self):
            return iter(self.zones.values())

# Foundation.

#
## Start generic.
# Note: this shares code with .runtime.architecture.Object, but there's still some
# reason to keep this package independent of the rest of stuphos.
class Object:
    class Meta:
        __attributes__ = ()

        @staticmethod
        def formatAttribute(instance, a):
            if type(a) in (list, tuple):
                if len(a) == 2:
                    return '%s = %r' % (a[0], getattr(instance, a[1]))
                if len(a) == 3:
                    return '%s = %r' % (a[0], getattr(instance, a[1], a[2]))
            elif type(a) is str:
                return '%s = %r' % (a, getattr(instance, a))

        @staticmethod
        def className(instance):
            return instance.__class__.__name__

        @classmethod
        def instanceRepr(self, instance):
            meta = instance.Meta
            attribs = ', '.join(meta.formatAttribute(instance, a) for \
                                a in meta.__attributes__)
            if attribs:
                return '<%s %s>' % (meta.className(instance), attribs)

            return '<%s>' % meta.className(instance)

    def __repr__(self):
        return self.Meta.instanceRepr(self)

    __str__ = __repr__

class Initializer(Object):
    class Meta(Object.Meta):
        __attributes__ = Object.Meta.__attributes__ + ('__dict__',)

    # Pattern encapsulates a dict within an attributable instance.
    def __init__(self, dict = None, **kwd):
        if dict is None:
            dict = {}

        dict.update(kwd)
        self.__dict__ = dict
    def update(self, **dict):
        self.__dict__.update(dict)
        return self # Method-chaining.

## End generic
#

class Record(Initializer):
    '''
    Pattern allows one of the world libdata reader classes
    to initialize this object, using lshift:

        class X(Record):
            def loadFrom(self, code):
                reader = RecordReader('x-%s.rec' % code)
                return self.genericLoadFromReader(reader)

        x = (X() << 'info')

    '''
    def __lshift__(self, *args, **kwd):
        self.loadFrom(*args, **kwd)

        # In-place operation; Method-chaining.
        return self

    def loadFromReader(self, reader):
        reader.parseFile()
        self.__dict__.update(reader.consumeRecordSet())

    # Default Behavior.
    loadFrom = loadFromReader

    ReaderClass = None
    def loadFromFile(self, file):
        return self.loadFromReader(self.ReaderClass(file))

# Entities.
class ZoneBase(Record):
    class ZCommand(Initializer):
        class Meta(Object.Meta):
            __attributes__ = ['type', 'conditional', 'arg1', 'arg2', 'arg3']

        class Arg:
            path = name = vnum = ''

        class EntityArg(Arg):
            @staticmethod
            def capfirst(s):
                return s[0].upper() + s[1:].lower()

            def __init__(self, world, vnum, type, **kwd):
                self.vnum = vnum
                self.path = '%s/%s' % (type.lower(), vnum)

                lookup = 'lookup' + self.capfirst(type)
                lookup = getattr(world, lookup)
                lookup = lookup(vnum)

                self.entity = lookup or ''
                self.name = getattr(lookup, 'name', '')

                self.__dict__.update(kwd)

        class MaxExistArg(Arg):
            def __init__(self, max):
                self.maximum = max
                self.name = '%s maximum' % max

        class WearPosArg(Arg):
            def __init__(self, world, pos):
                    self.position = pos
                    self.name = wearPosition(pos)

        class DirArg(Arg):
            def __init__(self, world, dir):
                    self.direction = dir
                    self.name = directionName(dir)

        class DoorArg(Arg):
            def __init__(self, world, state):
                    self.doorState = state
                    self.name = doorState(state)

        NullArg = Arg()

        def __init__(self, cmdtype, conditional, args, **kwd):
            arg1 = args[0]
            arg2 = args[1]
            arg3 = args[2] if len(args) > 2 else None

            Initializer.__init__(self,
                                 type = cmdtype,
                                 conditional = conditional,
                                 arg1 = arg1,
                                 arg2 = arg2,
                                 arg3 = arg3,
                                 **kwd)

        @property
        def name(self):
            return {'M': 'Load Mobile',
                    'O': 'Load Item',
                    'G': 'Give Item to Mobile',
                    'E': 'Equip Mobile',
                    'P': 'Put Item in Item',
                    'R': 'Remove Item from Room',
                    'D': 'Set Door State',
                    'U': 'Mount Mobile'} \
                    .get(self.type, "<Unknown '%c'>" % self.type)

        @property
        def args(self):
            'Return a resolved form of the arguments based on command type.'
            world = self.zone.world

            if self.type == 'M':
                    return [self.EntityArg(world, self.arg1, 'mobile'),
                            self.EntityArg(world, self.arg3, 'room'),
                            self.MaxExistArg(self.arg2)]

            if self.type == 'O':
                    return [self.EntityArg(world, self.arg1, 'item'),
                            self.EntityArg(world, self.arg3, 'room'),
                            self.MaxExistArg(self.arg2)]

            if self.type == 'G':
                    return [self.EntityArg(world, self.arg1, 'item'),
                            self.NullArg,
                            self.MaxExistArg(self.arg2)]

            if self.type == 'E':
                    return [self.EntityArg(world, self.arg1, 'item'),
                            self.WearPosArg(world, self.arg3),
                            self.MaxExistArg(self.arg2)]

            if self.type == 'P':
                    return [self.EntityArg(world, self.arg1, 'item'),
                            self.EntityArg(world, self.arg3, 'item'),
                            self.MaxExistArg(self.arg2)]

            if self.type == 'R':
                    return [self.EntityArg(world, self.arg2, 'item'),
                            self.EntityArg(world, self.arg1, 'room'),
                            self.NullArg]

            if self.type == 'D':
                    return [self.EntityArg(world, self.arg1, 'room'),
                            self.DirArg(world, self.arg2),
                            self.DoorArg(world, self.arg3)]

            if self.type == 'U':
                    return [self.EntityArg(world, self.arg1, 'mobile'),
                            self.EntityArg(world, self.arg3, 'mobile'),
                            self.MaxExistArg(self.arg2)]

            return [NullArg] * 3

    def __init__(self, world, vnum):
        """
        First argument is the base lib/world path.

        Second argument is the virtual no of this zone.  Note that zones can span
        multiple virtual numbers for room, object, and mobile sets.

        """
        Initializer.__init__(self)
        self.world = world

        # Load this record.
        (self) << (vnum)

        # Initialize further records.
        self.roomMap = {}
        self.mobileMap = {}
        self.itemMap = {}

    # Implementation of Record.
    ReaderClass = ZoneReader
    def loadFrom(self, vnum):
        return self.loadFromFile(self.world.fileBase.openZoneFile(vnum))

    # Properties.
    @property
    def resetModeName(self):
        return resetModeName(getattr(self, 'reset-mode'))

    @property
    def span(self):
        v = int(self.vnum)
        t = int(self.top)/100

        if v == t:
            return v

        return (v, t)

    @property
    def spanName(self):
        s = self.span
        return (type(s) is int and '#%d' or 'Spanning [%d-%d]') % s

    def spanZones(self):
        # Generator
        s = self.span

        if type(s) is int:
            yield s
        else:
            for z in range(s[0], s[1]+1):
                yield z

    @property
    def flagNames(self):
        return zoneFlagNames(self.flags)

    @property
    def continentName(self):
        return continentName(self.continent)

    @property
    def numZCmds(self):
        c = getattr(self, 'commands', None)
        return c and len(c) or 0

    def __repr__(self):
        return '<Zone %s (%s) %d/%d/%d/%d cmds/rooms/objs/mobs, %s every %d>' % \
            (self.spanName, self.name, self.numZCmds,
            len(self.roomMap), len(self.itemMap), len(self.mobileMap),
            self.resetModeName, self.lifespan)

    def createZCommand(self, *args, **kwd):
        # Factory.
        # Note this doesn't currently call into world.
        return self.ZCommand(*args, **kwd)

    def createRoom(self, vnum, record):
        # Factory.
        return self.world.createRoom(vnum, self, record)

    def createItem(self, vnum, record):
        # Factory.
        return self.world.createItem(vnum, self, record)

    def createMobile(self, vnum, record):
        # Factory.
        return self.world.createMobile(vnum, self, record)

    def loadRooms(self):
        # Consume, load and update entire room record set:
        reader = RoomReader(self.world.fileBase.openRoomFile(self.vnum))
        Record(self.roomMap) << (reader)

        # Reload room entities handles.
        for (vnum, room) in self.roomMap.items(): # XXX Dict change exception??
            self.roomMap[vnum] = self.createRoom(vnum, room)

        return self # Method-Chaining.

    def loadMobiles(self):
        # Consume, load and update entire mobile prototype record set:
        reader = MobileReader(self.world.fileBase.openMobileFile(self.vnum))
        Record(self.mobileMap) << (reader)

        # Reload mobile entities handles.
        for (vnum, mob) in self.mobileMap.items(): # XXX Dict change exception??
            self.mobileMap[vnum] = self.createMobile(vnum, mob)

        return self # Method-Chaining.

    def loadItems(self):
        # Consume, load and update entire object prototype record set:
        reader = ObjectReader(self.world.fileBase.openObjectFile(self.vnum))
        Record(self.itemMap) << (reader)

        # Reload item entities handles.
        for (vnum, item) in self.itemMap.items(): # Dict change exception??
            self.itemMap[vnum] = self.createItem(vnum, item)

        return self # Method-Chaining.

    @property
    def zcommands(self):
        return [self.createZCommand(*zcmd, **{'zone': self}) \
                for zcmd in getattr(self, 'commands', [])]

    @property
    def rooms(self):
        # Ascending sort by vnum.
        items = list(self.roomMap.items())
        items.sort()

        # Undecorate.
        return [i[1] for i in items]

    @property
    def items(self):
        # Ascending sort by vnum.
        items = list(self.itemMap.items())
        items.sort()

        # Undecorate.
        return [i[1] for i in items]

    @property
    def mobiles(self):
        # Ascending sort by vnum.
        items = list(self.mobileMap.items())
        items.sort()

        # Undecorate.
        return [i[1] for i in items]

    def resolveRooms(self):
        'Resolve room exits.'
        # XXX Is this needed?  Use lookupRoom on roomlink.
        raise NotImplementedError

    # def resolve(self):pass

class Room(Initializer):
    "['sector', 'name', 'exits', 'text', 'flags', 'descr', 'vnum']"

    class Meta(Object.Meta):
        __attributes__ = ['vnum', 'name', 'sector', 'flags']

    NOWHERE = -1

    class Exit(Initializer):
        "['room-link', 'keyword', 'descr', 'key', 'flags']"

        class Meta(Object.Meta):
            __attributes__ = ['room-link', 'key', 'flags', 'keyword']

        @property
        def roomlink(self):
            return getattr(self, 'room-link')

        @property
        def name(self):
            return directionName(self.dir)

        @property
        def isPurged(self):
            return (self.roomlink == Room.NOWHERE)

    @property
    def flagNames(self):
        return roomFlagNames(self.flags)

    @property
    def sectorName(self):
        return roomSectorName(self.sector)

    def createExit(self, *args, **kwd):
        # Factory.
        # Note this doesn't currently call into world.
        return self.Exit(*args, **kwd)

    @property
    def exitdirs(self):
        try: exits = self.exits
        except AttributeError: pass
        else:
            for (dir, info) in exits.items():
                yield self.createExit(info, dir = dir)

    def __repr__(self):
        return '<Room #%5d:%.50s (%s)>' % (self.vnum, self.name,
                                           ', '.join(self.flagNames))

class Mobile(Initializer):
    class Meta(Object.Meta):
        __attributes__ = ['vnum', 'shortdescr']

class Item(Initializer):
    """
    ['extraflags', 'actiondescr', 'longdescr', 'name',      'weight', 'antiflags', 'affections',
     'shortdescr', 'minlevel',    'cost',      'wearflags', 'vals',   'vnum',      'type',
     'extradescr', 'timer',       'trap',      'specproc']
    """

    class Meta(Object.Meta):
        __attributes__ = ['vnum', 'shortdescr', 'typeName']

    item_types = [] # Todo

    @property
    def typeName(self):
        if type(self.type) is int:
            if self.type >= 0 and self.type < len(self.item_types):
                return self.item_types[self.type]
        return 'Type%s' % self.type

    extraFlagNamesDict = {} # Todo
    wearFlagNamesDict = {} # Todo
    antiFlagNamesDict = {} # Todo

    @property
    def extraFlagNames(self):
        return flagNames(self.extraFlagNamesDict, self.extraflags)
    @property
    def wearFlagNames(self):
        return flagNames(self.wearFlagNamesDict, self.wearflags)
    @property
    def antiFlagNames(self):
        return flagNames(self.antiFlagNamesDict, self.antiflags)

    @property
    def value1(self):
        return self.vals[0]
    @property
    def value2(self):
        return self.vals[1]
    @property
    def value3(self):
        return self.vals[2]
    @property
    def value4(self):
        return self.vals[3]

# Utility.
class WorldLoader(WorldBase, WorldBase.Mixin):
    'Encapsulate all zon/wld/obj/mob/shp/sch data into accessible object.'

    def __init__(self, fileBase, absindex = None, logStream = False):
        self.fileBase = fileBase
        self.absindex = absindex
        self.logStream = logStream

        self.zoneMap = {}

    def logMessage(self, msg):
        if self.logStream:
            self.logStream.write(msg)
            self.logStream.flush()

    # Sequential DB load.
    def readZoneIndex(self, index = INDEX_FILE):
        # Generates zone vnums from zone index file parameters.
        for (vnum, fix) in self.fileBase.iterindex(self.fileBase.superindex(index, fix = Zonefix)):
            yield vnum

    def createZone(self, vnum):
        # Factory.
        return ZoneBase(self, vnum)

    def createRoom(self, vnum, zone, record):
        # Factory.
        return Room(record, vnum = vnum, zone = zone)

    def createItem(self, vnum, zone, record):
        # Factory.
        return Item(record, vnum = vnum, zone = zone)

    def createMobile(self, vnum, zone, record):
        # Factory.
        return Mobile(record, vnum = vnum, zone = zone)

    def loadWorld(self, index = INDEX_FILE, cascade = False):
        '''
        Cascading Circle-like database load from zone index.
        Loads all zones from specified index absolute path or
        base.  If cascade is True, load all other elements
        that are kept under zone (which is pretty much everything).
        '''

        self.logMessage('Loading index %r from %r:\n' % (index, self.fileBase))
        for vnum in self.readZoneIndex(index = index):
            self.loadZone(vnum, cascade = cascade)

    loadFromZoneIndex = loadWorld

    def loadZone(self, vnum, cascade = True):
        self.logMessage('    Info on Zone #%3d...' % vnum)
        zone = self.zoneMap[vnum] = self.createZone(vnum)

        if cascade:
            self.logMessage('Rooms...')
            zone.loadRooms()

            self.logMessage('Items...')
            zone.loadItems()

            self.logMessage('Mobiles...')
            zone.loadMobiles()

        self.logMessage('\n')
        return zone

    @property
    def zones(self):
        # Ascending sort by vnum.
        items = list(self.zoneMap.items())
        items.sort()

        # Undecorate.
        return [i[1] for i in items]

    def lookupZone(self, vnum):
        return self.zoneMap.get(int(vnum))

    def lookupRoom(self, vnum):
        # This should be done sooner, too.
        vnum = int(vnum)
        try: return self.zoneMap[vnum / 100].roomMap[vnum]
        except KeyError:
            return None

    def lookupMobile(self, vnum):
        vnum = int(vnum)
        try: return self.zoneMap[vnum / 100].mobileMap[vnum]
        except KeyError:
            return None

    def lookupItem(self, vnum):
        vnum = int(vnum)
        try: return self.zoneMap[vnum / 100].itemMap[vnum]
        except KeyError:
            return None

## Unused:
class WorldScanner(WorldBase):
    # Alternative load methods -- Random Access.
    def scanStart(self):
        pass
    def scanComplete(self):
        pass

    def scanParts(self, **parts):
        'Adaptive load from indexed sources.  Serve parts back to caller.'

        # Generator
        from types import GeneratorType as generator
        g = self.scanStart()

        if type(g) is generator:
            for n in g:
                yield n

        # Index record sources individually: by entity-set file unit or zone list.
        if 'rooms' in parts:
            # Scan room parts
            r = parts['rooms']

            if type(r) is str:
                index = self.superindex(index = r, fix = Roomfix)
            elif bool(r):
                # On truth test load default index.
                index = self.superindex(fix = Roomfix)

            for roomfile in self.iterindex(index):
                for r in self.scanRoomfile(roomfile):
                    yield r

        # Todo: provide objects
        # Todo: provide mobiles

        if 'zcommands' in parts:
            # Scan zone commands.
            pass

        if 'zone_table' in parts:
            # Scan all loadable zone data.
            pass

        if 'text':
            # Text search all text-aware locations.
            pass

        # End of scan.  Try to call this before exiting this function.
        g = self.scanComplete()
        if type(g) is generator:
            for n in g:
                yield n

    # Transformer.
    roomResult = staticmethod(lambda rec:rec) # Room(rec)

    def scanRoomfile(self, filename, sorted = False):
        rooms = RoomReader(filename).parseFile().consumeRecordSet()
        order = list(rooms.keys())

        if sorted:
            order.sort()

        for rec in rooms:
            #
            # Return an instance wrapping the data, which has an accessible 'universalForm' method.
            #
            yield self.roomResult(rec)

World = WorldLoader

# Front End.
def parse_cmdln(argv = None):
    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('--cascade', action = 'store_true')
    parser.add_option('-i', '--inspect', action = 'store_true')
    parser.add_option('-z', '--zone-index', '--index')
    parser.add_option('-l', '--log-file')
    return parser.parse_args(argv)

def buildWorld(options, args):
    '''
    Originally took argv and optional cascadeOptionName arguments.

    Now call like so:
        argv = ['']
        argv.append('/StuphMUD/lib/world')
        argv.append('--zone-index=index.mini')
        argv.append('--cascade')

        (options, args) = parse_cmdln(argv)
        w = buildWorld(options, args)

    This is basically what main() does, except validates # of arguments and
    implements optional interactive inspection.
    '''

    # Operates directly on 
    base = LocalIndex(args[0])
    absindex = False
    loadOptions = {}

    if options.zone_index:
        loadOptions['index'] = options.zone_index
        # absindex = isAbsIndex(options.zone_index)
        absindex = None # fs detect

    if options.cascade:
        loadOptions['cascade'] = True

    if options.log_file == '--':
        from sys import stdout as logging
    elif options.log_file:
        logging = open(options.log_file, 'w')
    else:
        logging = None

    w = World(base, absindex = absindex, logStream = logging)
    w.loadFromZoneIndex(**loadOptions)

    return w

def main(argv = None):
    (options, args) = parse_cmdln(argv)

    if not args:
        parser.print_usage()
        return

    w = buildWorld(options, args)
    if options.inspect:
        inspect(world = w)

def inspect(**kwd):
    from code import InteractiveConsole as IC
    import readline

    global pf, pp, page
    from pprint import pprint as pp, pformat as pf

    vars = globals()
    vars.update(kwd)
    IC(locals = vars).interact(banner = '')

if __name__ == '__main__':
    main()



# XMLRPC Host-Client Provision.
# todo: move this into ph.web/networking package?
from . import config, host, client

from .config import buildConfig
from .host import HostRpc, installMarshaller
from .client import ClientRpc

from stuphos.runtime.facilities import Facility
from stuphos import log as mudlog

class HostRpcManager(Facility, HostRpc):
    NAME = 'XMLRPC::Host'

    @classmethod
    def create(self):
        config = buildConfig(None)

        mudlog('Creating XMLRPC Endpoint (port %s)' % config.port)
        installMarshaller()

        methods = self.RpcMethods.get(create = True).methods
        return self(config, methods = methods)

    def __registry_delete__(self):
        self.stop()
        # Also, signal network . . .

    def __init__(self, *args, **kwd):
        HostRpc.__init__(self, *args, **kwd)
    def __str__(self):
        return '{XmlRpc Host}\n%s' % (self.RpcMethods.get(),)

    class Manager(Facility.Manager):
        VERB_NAME = 'xmlrpc-*host-manage'
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR

    class RpcMethods(Facility):
        NAME = 'XMLRPC::Methods'

        def __init__(self):
            self.methods = dict()
        def __str__(self):
            return '    ' + '\n    '.join(list(self.methods.keys()))

        ##    def _listMethods(self):
        ##        return []
        ##    def _methodHelp(self, method):
        ##        return []

HostRpcManager.manage()

# Classic interface.
def getHost(create = False):
    return HostRpcManager.get(create = create)
def delHost():
    return HostRpcManager.destroy()

def getClientProxy(config = None, **kwd):
    return ClientRpc(buildConfig(config), **kwd)

def HOSTCALL(name, heartbeat = False, result = None):
    # XXX Registered methods will not survive host reload!
    # XXX This is a very non-standard interface.
    def registerRpcMethod(function):
        hot = getHost()
        if heartbeat:
            host.register_heartbeat_function(function, result, name = name)
        else:
            host.register_function(function, name = name)
        return function
    return registerRpcMethod



'Base classes and utility functions for reading StuphMUD (Circle) world files.'

# Written by Fraun
# Todo: Rename RecordReader to RecordSetReader, derived from RecordReader which
#    only defines access of a single record.  Override parseFile as appropriate.

MINDEX_FILE  = 'index.mini'
INDEX_FILE   = 'index'
RINDEX_FILE  = 'index.reduced'

Zonefix = 'zon'
Roomfix = 'wld'
Objfix  = 'obj'
Mobfix  = 'mob'

class FormatError(Exception):
    def __init__(self, reason, cause = None, lineno = None):
        Exception.__init__(self, reason)
        self.cause = cause
        self.lineo = lineno

def ASCIIFlagDecode(flagstr):
    """Return packed integer bitvector representation of flag string as used in Circle MUD.

    bitvector_t asciiflag_conv(char *flag)
    {
          bitvector_t flags = 0;
          int is_num = TRUE;
          char *p;

          for (p = flag; *p; p++) {
            if (islower(*p))
              flags |= 1 << (*p - 'a');
            else if (isupper(*p))
              flags |= 1 << (26 + (*p - 'A'));

            if (!isdigit(*p))
              is_num = FALSE;
          }

          if (is_num)
            flags = atol(flag);

          return (flags);
    }
    """

    # Start bitvector value buffer at 0
    # For each character, set the nth bit, where n is:
    #    For lowercase characters, n is the offset into the alphabet (zero-based)
    #    For the uppercase, n is is the offset into the alphabet (zero-based), plus 26
    # (This makes valid bit-chars: a-zA-F)
    # Ignore all other characters
    # If all characters are digits, convert to long and that's the result
    # Otherwise, char-to-bit conversion is done, the value buffer is the result

    lowndx='abcdefghijklmnopqrstuvwxyz'.find # Find is faster: skips 1 instruction
    highndx='ABCDEF'.find

    bits=0
    numval=None # Need three values here

    for c in flagstr:
        if c.islower():
            bits|=1<<lowndx(c) # Guaranteed
        elif c.isupper():
            i=highndx(c)
            if i>-1:
                bits|=1<<(i+26) # Use long to suppress signage warning

        if not c.isdigit():
            numval=False
        elif numval is None:
            numval=True # Only occurs when there is at least a single digit

    if bool(numval): # False if None (no characters)
        return int(flagstr)

    return bits

def ASCIIFlagEncode(bits):
    flagstr=''
    for i in range(25):
        b=1<<i
        if bits&b==b:
            flagstr+='abcdefghijklmnopqrstuvwxyz'[i]

    for i in range(6):
        b=1<<(i+26)
        if bits&b==b:
            flagstr+='ABCDEF'[i]

    return flagstr

def flagNames(g, v):
    'sprintbit equivilant.'
    n = []
    a = n.append

    for x in list(g.keys()):
        b = (1 << x)
        if (v & b) == b:
            a(g[x])

    return n

# from string import printable
printable='~`!1@2#3$4%5^6&7*8(9)0_-+=QqWwEeRrTtYyUuIiOoPp{[}]AaSsDdFfGgHhJjKkLl:;"\'ZzXxCcVvBbNnMm,,>.?/|\\'

# Make this puppy printable!
def makePrintable(c):
    'An invalid file might contain binary data so this prevents a spam.'
    return (c in printable) and c or ('\\x%02x' % c)

import re
DICE_PATTERN = re.compile('([0-9\-]+)d([0-9\-]+)\+([0-9\-]+)')

def parseDice(dice):
    match = DICE_PATTERN.match(dice)
    assert match is not None, FormatError('Expected dice format (XdX+X) got: %r' % dice)
    return match.groups()

def isFileLike(file):
    for attr in ['read', 'readline']:
        if hasattr(file, attr):
            return True

    return False

class FileReader:
    libPath='lib'
    pathSep='/'

    def __init__(self, file):
        # from types import FileType
        from io import FileIO as FileType

        # Will likely be already opened by the wldlib.filesystem.
        fileType = type(file)

        # Want a file type
        if fileType is bytes:
            self.filename = file
            self.file = open(file)
        elif fileType is FileType:
            self.file = file
            self.filename = file.name
        elif fileType is int:
            # Shouldn't this be self.pathSep?
            self.filename = libPath + pathSep + str(file) + '.' + self.__class__.fileExtension
            self.file = open(self.filename)
        elif isFileLike(file):
            self.filename = file.name
            self.file = file
        else:
            raise TypeError(str(getattr(fileType, '__name__', '?')))

        self.lineno=0

    # Override this for other input methods
    def readLine(self):
        self.lineno+=1
        # todo: strip carriage returns.
        return self.file.readline() # .strip()

    def readLineNoEOF(self, errproc):
        self.lineno+=1

        n = self.file.readline() # .strip()
        if n != '':
            return n

        ##  if type(errproc) is str:
        ##      raise self.error(errproc)

        raise self.error(str(errproc))

    def error(self, msg):
        return FormatError('%s (%s:line %d)' % (msg, self.filename, self.lineno),
                           lineno = self.lineno)

    def tildeString(self, area='no particular'):
        'Read lines from file until terminated by "~\n", removing this sequence.'

        # Todo: utilize ReadTildeString from textlib.
        buffer=[]
        append=buffer.append

        while 1:
            line=self.readLine()
            assert line != '', self.error('Area: %s [EOF expecting ~string]' % area)

            # XXX should not strip every line (is this done in C source??)
            line=line.strip()
            tildeNdx=line.find('~')
            last=len(line)-1

            if tildeNdx > -1:
                if tildeNdx == last and tildeNdx != 0:
                    line=line[:-1]
                    append(line)

                break

            else:
                append(line)


        # There is a tilde: put it on, break on ~\n
        buffer='\n'.join(buffer)

        return buffer

    def readInteger(self):
        return int(self.readLine().strip())

    def readFloat(self):
        return float(self.readLine().strip())

    # Introduced by shops: (which use the v3format)
    def readNumberList(self):
        nItems=self.readInteger()

        numbers=[];append=numbers.append

        if getattr(self, 'v3format'):
            # This is Circle 3.0 shops.
            while True:
                # Fraun Jan 4th 2006 - Convert to readFloat (from readInteger)
                item=self.readFloat()
                if item<0.0:
                    break

                append(item)

        else:
            while nItems>0:
                nItems-=1
                append(self.readInteger())

        return numbers

# Multiple record reading.
class RecordProducer:
    'Interface.'

    VirtualNRName = 'vnum'
    nameOfRecordSet = 'recordSet'
    nameOfCurrentRecord = 'current'

    # Beginning of records (start of file)
    def startOfRecords(self):
        pass

    # End of records (end of file)
    def endOfRecords(self):
        pass

    # Start of new current record
    def newRecord(self, vnum):
        pass

    # Delete result reference from this record and return it to caller
    def consumeRecordSet(self):
        pass

    produce = consumeRecordSet

    # New parsed field name=value pair for current record
    def recordField(self, name, value):
        pass

    def getCurrentRecord(self):
        pass

    def getRecordSet(self):
        pass

class RecordProducerStore(RecordProducer):
    'Default response to record production: Store in dictionary backend.'

    def startOfRecords(self):
        setattr(self, self.nameOfRecordSet, {})

    def endOfRecords(self):
        if hasattr(self, self.nameOfCurrentRecord):
            delattr(self, self.nameOfCurrentRecord)

    def newRecord(self, vnum):
        r = {self.VirtualNRName:vnum}
        setattr(self, self.nameOfCurrentRecord, r)
        getattr(self, self.nameOfRecordSet)[vnum] = r

        # Usable?
        # return r

    def consumeRecordSet(self):
        r = getattr(self, self.nameOfRecordSet)
        delattr(self, self.nameOfRecordSet)

        return r

    def getCurrentRecord(self):
        return getattr(self, self.nameOfCurrentRecord)

    def getRecordSet(self):
        return getattr(self, self.nameOfRecordSet)

    def recordField(self, name, value):
        getattr(self, self.nameOfCurrentRecord)[str(name)] = value

    def getVirtualNumber(self):
        return getattr(self, self.nameOfCurrentRecord)[self.VirtualNRName]

# Fraun Dec 24th 2005 - Renamed from FileReader2 to RecordReader.
class RecordReader(FileReader, RecordProducerStore):
    'Implements commonalities between the parsers connoting structure.'

    MALFORMED_FIRST_RECORD_FMT = 'Malformed first record header %c%r (expected #<Virtual NR>)'

    def EOFNoDollar(self):
        raise self.error('Expected $ (dollar) to end records at EOF')

    def parseFile(self):
        'Initiates the parse of that file initialized in the constructor.'
        line = self.readLineNoEOF('Empty File')
        c = line[0]

        if c == '#':
            self.startOfRecords()

            vnum = int(line[1:])
            while vnum != None:
                self.newRecord(vnum)
                vnum = self.parseRecord()

            self.endOfRecords()

        elif c != '$':
            raise self.error(self.MALFORMED_FIRST_RECORD_FMT % (c, line))
        else:
            # Simulate empty record set.
            self.startOfRecords()
            self.endOfRecords()

        return self # Method chaining.

    def parseRecord(self):
        'Override this.'

    def readNextVnum(self):
        line = self.readLineNoEOF('Expected #<vnum> of next record').strip()
        if line[:1] == '$':
            return None

        assert line[:1] == '#', self.error('Expected #<vnum> of next record; got %r' % line[1:])

        try: return int(line[1:])
        except ValueError:
            raise self.error('Expected #<vnum> of next record; got %r' % line[1:])

class SingleRecordReader(RecordReader):
    # XXX how-the-fuck do I do this?
    pass

# Index File System.
def iterindex(index):
    for n in index:
        n = n.strip()
        if n == '$':
            break

        n = n.split('.')
        n[0] = int(n[0])
        yield n

def isAbsIndex(index):
    return (len(index) > 0 and (index[0] == '/')) or (len(index) > 1 and (index[1] == ':'))

class Index:
    'The general file-access routines, meant to abstract away from disk.'
    # FileBase: Merges FileSystem, Source/Store and Index.

    # Low-Level.
    def open(*args, **kwd):
        raise NotImplementedError
    def joinpath(*args, **kwd):
        raise NotImplementedError

    # Mid-Level.
    def fileName(self, vnum, fix):
        return self.joinpath(fix, '%d.%s' % (vnum, fix))

    # High-Level -- Index.
    def detectAbsIndex(self, index):
        return isAbsIndex(index)

    def superindex(self, index = INDEX_FILE,
                   absindex = False, fix = None):

        if absindex is None:
            absindex = self.detectAbsIndex(index)

        if not absindex:
            assert fix
            index = self.joinpath(fix, index)

        elif fix:
            index = self.joinpath(fix, index)

        return index

    def iterindex(self, index):
        return iterindex(self.open(index))

    def zonfile(self, vnum):
        return self.fileName(vnum, Zonefix)
    def roomfile(self, vnum):
        return self.fileName(vnum, Roomfix)
    def objfile(self, vnum):
        return self.fileName(vnum, Objfix)
    def mobfile(self, vnum):
        return self.fileName(vnum, Mobfix)

    def openZoneFile(self, vnum):
        return self.open(self.zonfile(vnum))
    def openRoomFile(self, vnum):
        return self.open(self.roomfile(vnum))
    def openObjectFile(self, vnum):
        return self.open(self.objfile(vnum))
    def openMobileFile(self, vnum):
        return self.open(self.mobfile(vnum))

class LocalIndex(Index):
    from os.path import join as joinpath
    joinpath = staticmethod(joinpath)

    def __init__(self, root):
        self.root = root

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self.root)

    def open(self, path, *args, **kwd):
        # Open relative from root.
        # XXX this breaks notions of absindex
        # Todo: collapse/normpath this.
        return open(self.joinpath(self.root, path), *args, **kwd)

# Archive Index Systems.
from io import StringIO

class ArchiveIndex(Index):
    # Build routines:
    @classmethod
    def openBuffer(self, buffer, filename = None):
        buffer.filename = filename
        buffer.name = filename # For ZipFile

        return self(buffer)

    @classmethod
    def openFieldStorage(self, storage, filename = None):
        return self.openBuffer(StringIO(storage.value),
                               filename = filename)

# Archive Provisions.
import zipfile

class ZipArchiveIndex(ArchiveIndex):
    # XXX prefix might not be needed anymore.
    def __init__(self, base):
        self.zipfile = zipfile.ZipFile(base)
        self.prefix = self.zipfile.filename + '/'
        self.prefix_len = len(self.prefix)

    def open(self, name):
        if name.startswith(self.prefix):
            name = name[self.prefix_len:]

        buf = StringIO(self.zipfile.read(name))
        buf.name = '%s%s' % (self.prefix, name)
        return buf

    def joinpath(self, *parts):
        return '/'.join(parts)

class TarArchiveIndex(ArchiveIndex):
    # XXX Impl.
    pass

_archive_index_typemap = dict(zip   = ZipArchiveIndex,
                              tar   = TarArchiveIndex,
                              gz    = TarArchiveIndex,
                              gzip  = TarArchiveIndex,
                              bz2   = TarArchiveIndex,
                              bzip2 = TarArchiveIndex)

def getFileExtension(name):
    # os.path.splitext is platform-dependent.
    i = name.rfind('.')
    if i >= 0:
        if name.rfind('/') < i:
            # (within this basename)
            return name[i+1:]

    return ''

def detectArchiveIndexType(base):
    typeName = getFileExtension(base).lower()
    return _archive_index_typemap.get(typeName)

# Factories.
##    def buildArchiveIndex(base, filename = None):
##        fstype = detectArchiveIndexType(filename or base)
##        if fstype is not None:
##            return fstype(base, filename = filename)

# XXX Needs Compat:
#   Now, using detectArchiveIndexType, obtain indexClass
#   and then call appropriate build routines.


# Routines.
def IsBitSet(vector, bit):
    return bool(vector & bit)
def SetBit(vector, bit):
    return (vector | bit)
def RemoveBit(vector, bit):
    return (vector | ~(bit))

def getFlagBit(flags, name):
    # Eww..
    flags = list(flags.values())
    lname = name.lower()

    for bit in range(len(flags)):
        if flags[bit].lower() == lname:
            return bit

    raise NameError(name)

def getFlagVector(flags, name):
    return (1 << getFlagBit(flags, name))

def IsFlagSet(vector, flags, name):
    return IsBitSet(vector, getFlagVector(flags, name))
def SetFlag(vector, flags, name):
    return SetBit(vector, getFlagVector(flags, name))
def RemoveFlag(vector, flags, name):
    return RemoveBit(vector, getFlagVector(flags, name))

from .constants import _level_names, zoneFlagNamesDict, _reset_mode_names, _continents
from .constants import _directions, _room_sectors, _movement_loss, roomFlagNamesDict
from .constants import _door_states, _wear_positions

def GetLevelName(level):
    try: return _level_names[level]
    except KeyError:
        raise ValueError(level)

def GetSectorMoveCost(sector1, sector2):
    # move points needed is avg. move loss for src and destination sect type
    return (_movement_loss.get(sector1, 0) + \
            _movement_loss.get(sector2, 0)) / 2

def zoneFlagNames(flags):
    return flagNames(zoneFlagNamesDict, flags)

def resetModeName(reset_mode):
    return _reset_mode_names.get(reset_mode) or \
           ('Reset Mode #%s' % reset_mode)

def continentName(continent):
    return _continents.get(continent) or \
           'Continent #%s' % continent

def directionName(dir):
    return {0: 'north', 1: 'east', 2: 'south',
            3: 'west',  4: 'up',   5: 'down'} \
            .get(dir) or 'Unknown: %s' % dir

def directionIndicator(dir):
    try: return _directions[int(dir)]
    except (IndexError, ValueError):
        return ''

def directionByName(dir):
    return {'north': 0, 'east': 1, 'south': 2,
            'west' : 3, 'up'  : 4, 'down' : 5}.get(str(dir))

def numberOfDirections():
    return len(_directions)

def doorState(state):
    return _door_states.get(state) or 'Door State: %s' % state

def wearPosition(pos):
    return _wear_positions.get(pos) or 'Wear Position: %s' % pos

def roomFlagNames(flags):
    return flagNames(roomFlagNamesDict, flags)

def roomSectorName(sector):
    return _room_sectors.get(sector) or 'Sector #%s' % sector

from .constants import _preference_bits, _preference_bits2
def allPreferenceFlagNames(vector1, vector2):
    return flagNames(_preference_bits, vector1) + \
           flagNames(_preference_bits2, vector2)

def isPreferenceFlagSet(vector1, vector2, name):
    try: return isPreference1FlagSet(vector1, name)
    except NameError:
        return isPreference2FlagSet(vector2, name)

def isPreference1FlagSet(vector, name):
    return IsFlagSet(vector, _preference_bits, name)
def isPreference2FlagSet(vector, name):
    return IsFlagSet(vector, _preference_bits2, name)

def preference1FlagNames(vector):
    return flagNames(_preference_bits, vector)
def preference2FlagNames(vector):
    return flagNames(_preference_bits2, vector)

def setPreference1Flag(vector, name):
    return SetFlag(vector, _preference_bits, name)
def setPreference2Flag(vector, name):
    return SetFlag(vector, _preference_bits2, name)

def removePreference1Flag(vector, name):
    return RemoveFlag(vector, _preference_bits, name)
def removePreference2Flag(vector, name):
    return RemoveFlag(vector, _preference_bits2, name)



# Client Provision.
from xmlrpc.client import ServerProxy
import socket

CLIENT_TIMEOUT = 20
CLIENT_URL_FORMAT = 'http://%(hostname)s:%(port)d/RPC2'

def getClientUrl(hostname, port):
    return CLIENT_URL_FORMAT % dict(hostname = hostname,
                                    port = port)

class ClientRpc:
    def __init__(self, config, timeout = CLIENT_TIMEOUT, verbose = False):
        # This should be configurable via client command-line.
        self.timeout = timeout
        self.proxy = ServerProxy(getClientUrl(config.hostname,
                                              config.port),

                                 verbose = verbose)

    def __call__(self, name, *args, **kwd):
        proc = getattr(self.proxy, name, None)

        try: return doSettingTimeout(self.timeout, proc, *args, **kwd)
        except socket.error as e:
            if e.args[0] != ECONNREFUSED:
                raise

def doSettingTimeout(timeout, function, *args, **kwd):
    # Dirty hack?  Presumably the rpc call is when the socket is made.
    if timeout is None:
        return function(*args, **kwd)

    previous = socket.getdefaulttimeout()
    socket.setdefaulttimeout(timeout)
    try: return function(*args, **kwd)
    finally: socket.setdefaulttimeout(previous)



# Socket Programming & Timing.
from stuphos.system import Heartbeat
from stuphos.system.core import getSystemTime

import socket
import asyncore
import asynchat
import queue

from errno import EINPROGRESS, EALREADY, EWOULDBLOCK, EISCONN

class MotherSocket(Heartbeat.Task, asyncore.dispatcher):
    BIND_ADDRESS = '0.0.0.0'
    PULSE_FREQUENCY = 0.1

    class Connection(asynchat.async_chat):
        def __init__(self, *args, **kwd):
            asynchat.async_chat.__init__(self, *args, **kwd)

        def handle_close(self):
            self.network.closeConnection(self)
            self.close()

    class IncomingConnection(Connection):
        def __init__(self, network, socket, addr):
            MotherSocket.Connection.__init__(self, sock = socket,
                                             map = network.socketMap)

            self.client_address = addr
            self.network = network

            from stuphos.system.api import game
            self.peer = game.new_peer(self) # XXX

            # This should be done in new_peer:
            self.peer.host = addr[0]
            self.peer.remote_port = addr[1]

            # Devel compatibility:
            self.peer.engine = network.engine

            # Configure chat.
            self.set_terminator('\n')
            self.buffer = ''

            # Internal Greetings/State Initialization.
            from stuphos import log as mudlog # not the same as game.mudlog..?
            mudlog('NEW CONNECTION: %r' % self)

            from stuphos.system.api import world
            self.peer.state = world.player.nanny.CON_GET_NAME

            # World Bridge.
            StuphMUD.NewIncomingPeer(self.peer)
            StuphMUD.GreetPlayer(self.peer)

        # Chat I/O.
        def collect_incoming_data(self, data):
            self.buffer += data
        def found_terminator(self):
            input = self.buffer.replace('\r', '')
            self.buffer = ""

            # todo: telnet processing -- XXX this doesn't work with terminator-based input.
            self.network.handlePlayerInput(self.peer, input)

    def __init__(self, port, address = None, frequency = None):
        self.socketMap = {}
        asyncore.dispatcher.__init__(self, map = self.socketMap)

        self.port = port
        self.address = address is None and self.BIND_ADDRESS or address

        self.frequency = frequency or self.PULSE_FREQUENCY
        self.last_pulse = None
        self.connections = []
        self.connect_q = queue.Queue()

    def activate(self, engine):
        if self.socket is None:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.set_reuse_addr()
            self.bind((self.address, self.port))
            self.listen(5)

            print('Activated %r' % self)

        engine.network = self
        self.engine = engine

    def deactivate(self, engine):
        if self.socket is not None:
            self.close()

        try:
            if engine.network is self:
                del engine.network
            if self.engine is engine:
                del self.engine
        except AttributeError:
            pass

    def perform(self, engine):
        # Todo: catch .nextTimeout => SystemError: Time moving backwards.
        # This seems to happen when the dev computer comes out of hibernation.
        try: timeout = self.nextTimeout
        except SystemError as e:
            print(f'{self.__class__.__name__}.perform: {e}')
        else:
            asyncore.loop(timeout = timeout,
                          map = self.socketMap,
                          count = 1)

            try:
                while True:
                    conn = self.connect_q.get_nowait()
                    conn.handle_connect()

            except queue.Empty:
                pass

    def getNextTimeout(self):
        now = getSystemTime()
        timeout = self.frequency
        if self.last_pulse is not None:
            # Calculate time remaining within pulse from last call.
            elapsed = now - self.last_pulse
            if elapsed < 0:
                raise SystemError('Time moving backwards! (%s)' % elapsed)

            # print 'elapsed network time:', elapsed
            self.last_pulse = now
            timeout = elapsed % self.frequency
            if not timeout:
                timeout = self.frequency

        self.last_pulse = now
        # print 'next timeout:', timeout
        return timeout

    # Override for different timeout mode.
    nextTimeout = property(getNextTimeout)

    ##    def __getattr__(self, name):
    ##        try: return self.__dict__[name]
    ##        except KeyError:
    ##            return asyncore.dispatcher.__getattr__(self, name)

    # Mother Socket Dispatcher.
    def handle_accept(self):
        (socket, addr) = self.accept()
        self.newConnection(socket, addr)

    def newConnection(self, socket, address = None):
        conn = self.IncomingConnection(self, socket, address)
        self.connections.append(conn)

    def closeConnection(self, peer):
        if peer in self.connections:
            # Error occurred before newConnection was called.
            # XXX This doesn't really make sense.. newConnection
            # should always be called (synchronously) before close...
            self.connections.remove(peer)
            StuphMUD.Disconnection(peer)

    class OutgoingConnection(Connection):
        def __init__(self, network, sock, address):
            MotherSocket.Connection.__init__(self, sock = sock,
                                             map = network.socketMap)

            if sock is None:
                self.create_socket(socket.AF_INET, socket.SOCK_STREAM)

            self.remote_address = address
            self.network = network

            # Configure chat.
            self.set_terminator('\n')
            self.buffer = ''

            from _thread import start_new_thread as nth
            nth(self.connectOpen, ())

        def connectOpen(self):
            self.connected = False
            err = self.socket.connect_ex(self.remote_address)
            # XXX Should interpret Winsock return values
            if err in (EINPROGRESS, EALREADY, EWOULDBLOCK):
                # C. Banis: For instance EINPROGRESS on cygwin is common for
                # a successful connection (0, EISCONN), meaning that the full
                # dispatcher cycle is never completed.  If this isn't done,
                # normal asyncore.dispatcher.handle_read_event will handle the
                # connect, but it will bypass our synchronization mechanism.
                err = 0 # return
            if err in (0, EISCONN):
                self.addr = self.remote_address
                self.connected = True
                self.network.handleOutgoingConnect(self)
            else:
                raise socket.error(err, errorcode[err])

        def handle_close(self):
            self.network.closeConnection(self)
            self.close()

        def handle_connect(self):
            print('Outgoing Connection Established [%s:%s]' % self.remote_address)

        # Chat I/O.
        def collect_incoming_data(self, data):
            self.buffer += data
        def found_terminator(self):
            self.network.handleIncomingData(self, self.buffer)
            self.buffer = ""

    def openConnection(self, address = None, socket = None):
        conn = self.OutgoingConnection(self, socket, address)
        self.connections.append(conn) # Do this aft-thread?
        return conn

    def handleOutgoingConnect(self, conn):
        self.connect_q.put(conn)
    def handlePlayerInput(self, peer, input):
        peer.forceInput(input)
        if self.engine.event:
            self.engine.event += peer.handleNextInput
    def handleIncomingData(self, conn, data):
        if self.engine.event:
            self.engine.event.call(handleIncomingData, conn, data)

    def emergeInternalPeer(self, stateName = None, avatar = None):
        ##    from game.cli import Console
        ##    return Console()

        from world import peer
        from world.player.nanny import CON_GET_NAME as DEFAULT_STATE

        new = peer.create()
        new.state = stateName or DEFAULT_STATE
        if avatar:
            new.avatar = avatar

        new.engine = self.engine

        return new

def handleIncomingData(conn, data):
    # Game Inline Event.
    print('%%', data)

class FreeNetwork(MotherSocket):
    class Engine:
        pass

    nextTimeout = None
    engine = Engine()

    def run(self):
        while self.is_running():
            self.perform(self.engine)

    def is_running(self):
        try: return self.__running
        except AttributeError:
            return False

    def set_running(self, state = True):
        self.__running = state

    def start(self):
        if not self.is_running():
            self.set_running(True)
            from _thread import start_new_thread as nth
            nth(self.run, ())

        return self

    def stop(self):
        self.set_running(False)
        return self

# Event Bridge Bindings.
StuphMUD.NewIncomingPeer = 'newConnection'
StuphMUD.Disconnection = 'disconnection'
StuphMUD.GreetPlayer = 'greetPlayer'



# Line-based In-Game String Content Editor
from types import GeneratorType as generator, MethodType
import re

# First:
def joinString(function):
    def doJoin(*args, **kwd):
        # (But not NL at end-of-block)
        return '\n'.join(function(*args, **kwd))

    return doJoin

def quote(s):
    return ' %  ' + '\n %  '.join(s.split('\n')) + '\n'

CLEAR_SCREEN = '\x1b[H\x1b[J'

def ParseRegExpr(line):
    # Todo: rewrite
    pattern = []
    escape = False
    for x in range(len(line)):
        c = line[x]
        if c == '\\':
            escape = True
            continue

        elif c == '/':
            if not escape:
                break

        escape = False
        pattern.append(c)

    pattern = ''.join(pattern)
    line = line[x+1:]

    i = line.rfind('/')
    if i >= 0:
        action = line[i+1:]
        line = line[:i]
    else:
        action = None

    yield (action, pattern, line)

class LineEditor:
    # Expansive 'state' stack and special command-handling for frameworking,
    # also, exclusive 'paging/viewing' mode.
    class State:
        MAIN = 0
        HELP = 1
        PASTE = 2
        VIEWING = 3

        INITIAL = MAIN

    HELP_SCREEN = CLEAR_SCREEN + "Commands:\n\n     @     Quit Editor               h ?   Help Screen\n" + \
                                              "     n     Next Line                 p     Previous Line\n" + \
                                              "     f     Forward Half Page         b     Backward Half Page \n" + \
                                              "     ff    Forward Full Page         bb    Backward Full Page\n" + \
                                              "     s     Start of Content          e     End of Content\n" + \
                                              "\n" + \
                                              "     --    Delete Current Line\n" + \
                                              "     ...   Begin Paste Mode\n" + \
                                              "\n" + \
                                              "     #<line content>     Change Current Line\n" + \
                                              "     +<line content>     Insert After Current Line\n" + \
                                              "     <<line content>     Insert Before Current Line\n" + \
                                              "     > j<line number>    Jump to Line Number\n" + \
                                              "\n" + \
                                              "     /<pattern>/<substitution>/p|g\n" + \
                                              "\n" + \
                                              "     ;<special command>  See Extended Documentation\n" + \
                                              "\n\n(Hit 'x' to exit help)\n"

    def __init__(self, peer, xxx_todo_changeme1, content = '', finishproc = None,
                 position = None, message = '', header = None, state = None):

        (rows, cols) = xxx_todo_changeme1
        assert isinstance(rows, int) and rows > 2
        assert isinstance(cols, int) and cols > 10

        self.content = self.loadContent(content)
        self.rows = rows
        self.cols = cols

        self.finishproc = finishproc
        self.position = 0 if position is None else self.validatePosition(position)

        self.header = header
        self.message = message
        self.state = []

        self.pushState(state if isinstance(state, int) else self.State.INITIAL)

        # Start prompting.
        PromptFor(peer, self.inputPhase, message = self.getScreen(),
                  write_mode = True, compact_mode = True)

    def loadContent(self, content):
        if isinstance(content, str):
            return content.split('\n')

        assert isinstance(content, (list, tuple, generator))
        return list(content)

    def renderContent(self):
        return '\n'.join(self.content) # + '\n'

    def validatePosition(self, pos):
        pos = int(pos)
        n = len(self.content)
        assert pos <= n and pos > 0, AssertionError('Valid positions are 1-%d' % n)
        return pos

    def setMessage(self, message):
        self.message = str(message)
    def setPrompt(self, prompt):
        # Note it's the prompt's message, not this editor's message.
        self.prompt.message = prompt

    def inputPhase(self, prompt, peer, line):
        self.prompt = prompt
        self.handleCommand(peer, line)

        if self.state:
            self.setPrompt(self.getScreen())
            return True # Regenerate

        if callable(self.finishproc):
            self.finishproc(self, peer)

        return False

    def inState(self, *states):
        return self.state[0] in states
    def pushState(self, state):
        self.state.insert(0, state)
    def popState(self):
        del self.state[0]

    @joinString
    def getScreen(self):
        if self.inState(self.State.PASTE):
            pass # Mainly an input phase.  But up here for optimization.

        elif self.inState(self.State.MAIN, self.State.VIEWING):
            r = self.rows - 1 # in between last line and message/prompt

            r -= 1 # our header
            yield CLEAR_SCREEN # ''

            if self.header:
                r -= 2
                yield self.header
                yield ''

            if self.message:
                r -= 1

            h = r // 2
            s = self.position - h
            e = self.position + h

            if s < 0:
                e -= s

            c = self.content
            n = len(c)
            if e >= n:
                s -= (e - n)
                e = n

            if s < 0:
                s = 0

            window = self.content[s:e]
            w = len(str(e))
            fmt = '%%s %%%dd   %%s' % w

            o = self.cols
            oMinus3 = o - 3
            for x in range(s, e):
                i = fmt % ('>' if x is self.position else ' ', x + 1, c[x])
                if len(i) > o:
                    i = i[:oMinus3] + '...'

                yield i

            if e < n:
                # ! This doesn't make much sense...
                yield (' [%%%dd]' % len(str(e))) % n

            yield ''
            if self.message:
                yield self.message

            yield '> '

        elif self.inState(self.State.HELP):
            yield self.HELP_SCREEN

    def handleCommand(self, peer, line):
        # Um, undo buffer??
        if self.inState(self.State.PASTE):
            if line == '@':
                self.pasteMode(False)
            else:
                self.pasteLine(line)

        elif self.inState(self.State.MAIN, self.State.VIEWING):
            line = line.strip()
            if not self.handlePaging(peer, line) and self.inState(self.State.MAIN):
                if line == '--':
                    self.delete(self.position)
                elif line == '...':
                    self.pasteMode(True)

                elif line:
                    c = line[0]
                    line = line[1:]

                    if c == '#':
                        # Edit/replace the line.
                        self.edit(self.position, line)

                    elif c == '+':
                        self.insert(self.position + 1, line)
                    elif c == '<':
                        self.insert(self.position, line)

                    elif c == '/':
                        # Search & Replace
                        (action, pattern, replace) = ParseRegExpr(line)
                        if action == 'p': # Current line.
                            pattern = re.compile(pattern)

                            line = self.content[self.position]
                            change = pattern.sub(line, replace)

                            self.edit(self.position, change)

                    elif c == ';': # '!'? -- this blocks all other interpreter action.
                        return self.specialCommand(peer, line)
                    else:
                        self.setMessage('NO: %s%s' % (c, line))

        elif self.inState(self.State.HELP):
            if line == 'x':
                self.popState()

        else:
            import ph
            ph.log('Unknown editor state (%s): %s' % (peer, self.state[0]))
            self.popState()

    def handlePaging(self, peer, line):
        if line == '@':
            self.popState()

        # Don't reset.
        # self.setMessage('')

        elif line == 'n':
            self.go(1)
        elif line == 'p':
            self.go(-1)
        elif line == 'f':
            self.go(self.rows // 2)
        elif line == 'b':
            self.go(-self.rows // 2)
        elif line == 'ff':
            self.go(self.rows)
        elif line == 'bb':
            self.go(-self.rows)
        elif line == 's':
            self.position = 0
        elif line == 'e':
            self.position = len(self.content) - 1
        elif line in ['h', 'help', '?']:
            self.pushState(self.State.HELP)
        elif not line and self.inState(self.State.VIEWING):
            # Just page -- like ff.
            self.go(self.rows)

        elif line[:1] in 'j>':
            try: self.jump(int(line[1:].strip()))
            except ValueError as e:
                self.setMessage(str(e))

        else:
            return False

        return True

    # Editing Methods.
    def go(self, nr = 0):
        self.position = min(max(self.position + nr, 0), len(self.content) - 1)
    def delete(self, pos):
        del self.content[self.position]

        n = len(self.content)
        if self.position >= n and n:
            self.position = n - 1

    def edit(self, pos, line):
        if self.content:
            self.content[self.position] = line
        else:
            # We maintain position 0 (line 1), even if it doesn't
            # exist.  So the special case is here, to create it.
            self.content.append(line)

    def insert(self, pos, line):
        if pos > len(self.content):
            self.content.append(line)
        elif pos >= 0:
            self.content.insert(pos, line)

        self.position += 1

    def pasteMode(self, state):
        if state:
            self.insertedNr = 0
            self.pushState(self.State.PASTE)
        else:
            self.setMessage('[Inserted %d lines]' % self.insertedNr)
            self.popState()

    def pasteLine(self, line):
        self.insert(self.position + 1, line)
        self.insertedNr += 1

    def jump(self, pos):
        try: self.position = self.validatePosition(pos) - 1
        except AssertionError as e:
            self.setMessage(e)

    def specialCommand(self, peer, command):
        self.setMessage('???: ;%s' % command)
        return True

class TreeEditor:
    pass # not yet implemented

# Invocation.
def getPeerWindowSize(peer):
    try: cols = peer.avatar.properties.screen.width
    except AttributeError:
        cols = 70

    rows = peer.avatar.page_length
    if rows <= 0:
        rows = 20
    elif rows < 5:
        rows = 5

    return (peer.avatar.page_length, cols)

def HeaderToFit(header, cols):
    n = cols - 6
    return ('&w[&m*&N %%%d.%ds &m*&w]&N' % (n, n)) % ('Editing: ' + header)

def TerminateMessageEditing(editor, peer):
    if peer.messenger:
        peer.messenger(peer, editor.renderContent())

def RichEditString(peer, content, header = None):
    (rows, cols) = getPeerWindowSize(peer)
    LineEditor(peer, (rows, cols), content,
               finishproc = TerminateMessageEditing,
               header = HeaderToFit(header, cols) \
                        if header else '')

def RichPageString(peer, content, header = None):
    (rows, cols) = getPeerWindowSize(peer)
    LineEditor(peer, (rows, cols),
               state = LineEditor.State.VIEWING,
               header = HeaderToFit(header, cols) \
                        if header else '')

def DecisiveEditString(peer, content, *args, **kwd):
    if getattr(peer.avatar, 'useRichEditor', False):
        return RichEditString(peer, content, *args, **kwd)

    return peer._basicEditString(content)

def DecisivePageString(peer, content, header = None):
    if getattr(peer.avatar, 'useRichPager', False):
        # Force needing header for now?  Otherwise, we end up paging
        # when we don't want to -- eventually, skip entering serious
        # paging if there's only one page (unless explicit).
        if header:
            return RichPageString(peer, content, header = header)

    return peer._basicPageString(content)

# thisModule = lambda:__import__(__name__, fromlist = [''])
def bindToThis(object, method):
    return MethodType(method, object)

def HookPeerEditor(peer):
    peer._basicEditString = peer.editString
    peer.editString = bindToThis(peer, DecisiveEditString)

def HookPeerPager(peer):
    peer._basicPageString = peer.page_string
    peer.page = peer.page_string = bindToThis(peer, DecisivePageString)

def Edit(peer, content = '', *args, **kwd):
    # (Parameterized) Decorator that starts the editing process.
    def startEditing(function):
        def terminateWithDisconnect(peer, content):
            peer.messenger = None
            return function(peer, content)

        peer.messenger = terminateWithDisconnect
        peer.editString(content, *args, **kwd)

        return None

    ##    def startEditing(function):
    ##        baseline = md5sum(content)
    ##        def saveIfChanged(peer, content):
    ##            if md5sum(content) != baseline:
    ##                function(peer, content) # prompt-for?
    ##            # else print something
    ##
    ##        peer.messenger = saveIfChanged
    ##        peer.editString(content, *args, **kwd)
    ##
    ##        return None

    return startEditing

# API.
@runtime.api(runtime.Player.Editor.API)
class EditorAPI:
    Edit = StartEditing = staticmethod(Edit)
    RichEditString = staticmethod(RichEditString)
    RichPageString = staticmethod(RichPageString)

# Extension Package.
from stuphmud.server.player import ACMD, ACMDLN, Option # For later.
from stuphos.runtime import Component

@ACMD('rich-editor*-select')
def doSelectRichEditor(peer, cmd, argstr):
    a = peer.avatar
    if a is not None:
        a.useRichEditor = not getattr(a, 'useRichEditor', False)
        print('You will now use the', \
                       'rich' if a.useRichEditor else 'basic', \
                       'text editor.', file=peer)

    return True

@ACMD('rich-pager*-select')
def doSelectRichPager(peer, cmd, argstr):
    a = peer.avatar
    if a is not None:
        a.useRichPager = not getattr(a, 'useRichPager', False)
        print('You will now use the', \
                       'rich' if a.useRichPager else 'basic', \
                       'string pager.', file=peer)

    return True

# Implement as component until it's moved into the core.  Then
# the hook install it can be called directly.
class RichText(Component):
    def onNewConnection(self, cltr, peer):
        try:
            HookPeerEditor(peer)
            HookPeerPager(peer)

        except Exception as e:
            print(f'[rich text:new connection] {e.__class__.__name__}: {e}')


@ACMDLN('libedit', Option('--list-directory', '--directory',
                          '--main', action = 'store_true'),
                   Option('--display', '--show',
                          action = 'store_true'))
def doEditLibrary(peer, command):
    from stuphmud.server.player import policy
    if peer.avatar in policy:
        from stuphlib import directory as stuphdir
        if command.options.list_directory:
            # Show main directory.
            peer.page(str(stuphdir.StuphLIB))

        elif command.args:
            object = command.args[0]

            try: filename = stuphdir.StuphLIB[object]
            except stuphdir.LibDataError as e:
                print('&r%s: %s&N' % (e.__class__.__name__, e), file=peer)
            else:
                if command.options.display:
                    with open(filename) as fl:
                        peer.page(fl.read())
                else:
                    # Edit the file.
                    with open(filename) as fl:
                        @Edit(peer, fl.read(), header = object)
                        def stuphlibObject(peer, content):
                            from shutil import copyfile
                            copyfile(filename, filename + '.bak')

                            with open(filename, 'wb') as fl:
                                fl.write(content)

                            print('Updated %s to %s (with backup)' % \
                                  (object, filename), file=peer)

        else:
            peer.page(command.help())

        return True


# Testing.
# is string.join(list|tuple) faster than string.join(generator)?
def pageLines(peer, s):
    def rebuild():
        x = s.split('\n')
        fmt = ' %%%dd  %%s' % len(str(len(x)))
        for (i, n) in enumerate(s.split('\n')):
            yield fmt % (i+1, n)

    peer.page_string('\n'.join(rebuild()) + '\n')

def main(argv = None):
    global PromptFor
    class PromptFor:
        def __init__(self, peer, callback, message = '', **kwd):
            self.callback = callback
            self.message = message

            peer.interpreter = self
            peer.prompt = message

        def __call__(self, peer, line):
            if self.callback(self, peer, line):
                self.regenerate(peer)
                return True

        def regenerate(self, peer):
            self.__class__(peer, self.callback, self.message)

    class Peer:
        def __init__(self, xxx_todo_changeme):
            (rows, cols) = xxx_todo_changeme
            self.rows = rows
            self.cols = cols
            self.prompt = None

        def interact(self):
            try:
                while self.interpreter(self, input(self.prompt)):
                    pass

            except EOFError:
                print() # self.interpreter(self, <quit-sequence>)

        def editString(self, content, header = None):
            edlin = LineEditor(self, (self.rows, self.cols), content,
                               header = HeaderToFit(header, self.cols) \
                               if header else '')
            self.interact()
            return edlin.content

    from optparse import OptionParser
    parser = OptionParser()
    parser.add_option('--rows', type = int, default = 20)
    parser.add_option('--columns', type = int, default = 100)
    (options, args) = parser.parse_args(argv)

    dimensions = (options.rows, options.columns)

    content = open(args[0]).read() if args else ''
    Peer(dimensions).editString(content, header = args[0])

if __name__ == '__main__': main()
else: from stuphmud.server.player.shell import PromptFor



# Synthetic Game Module
# --
# (C) 2021 runphase.com .  All rights reserved.
#

from .compartment import *

# These things need to be explicitly imported because of their usage through the application.
# To rely on game module for a particular implementation means that your application should
# upgrade to be using compartment.  In otherword, relying on game module except in a limited
# embedded context, is being obseleted.

# from compartment.world import new_peer

from .compartment.system import mudlog, boot_time, Heartbeat, Heartbeat as Engine
from .compartment.system import getCmdln, parseCmdln, Core, Game, mudlog as log
#from compartment.system.access import hasOLCAccessByName, iterateOLCAccessByName, shutdownMother
# from compartment.system.circle import mini_mud

# from compartment.misc import NanoText, registerRealPath, columnize
from .compartment.misc import *

# from compartment.world import peer, mobile_instance, player_character, getPlayerStoreByName, iteratePlayerIDs, findPlayerNameByID, SPECIAL
# from compartment.world.player import getMasterCommandList, interpretCommand, loadPlayer, EnableConsole
# from compartment.world.player import nanny, commands
# from compartment.world.player.commands import lookAtRoom

# from compartment.cli import Console


# The problem is that, if you're using this game module, we assume that you're using all of
# the intended functionality in the parallel package.  We do that individually here.

from .compartment import system
# from .compartment import world
# from .compartment import engine
from .compartment import format
from .compartment import misc
# from compartment import emulation


def iterateOLCAccessByName(name):
	return []



##    MIME-Version: 1.0
##    Content-Type: text/plain
##    X-Checkpoint-Command-Type: application/x-stuph-python-script

#X-Code-Source: http://www.penobscotrobotics.us/team/admin
#X-Code-Source: automatic
#X-Script-Name: penro.team.admin
#X-Script-Debug: on

# Incorporate a new componentized facility into the core.
from stuphos.runtime.facilities import Facility
from stuphos.runtime import Synthetic

class TeamCore(Facility):
    NAME = 'PenobscotRobotics::Team::Core'
    COMPONENT = True

    class Manager(Facility.Manager):
        VERB_NAME = 'penro-*team-core'
        # todo: commands that add/remove members!

    ##    def __init__(self):
    ##        self.members = # load from file/database

    # Watch the world:
    def onEnterGame(self, ctlr, peer, player):
        # todo: add this property only if it's a team member!!
        # player.properties.team = Synthetic(name = 'penro')

        player.team = Synthetic(name = 'penro',
                                core = self)

    def __str__(self):
        return '{Team Members: 0}'

if TeamCore.CreateNew(andManage = True):
    from stuphos.system.api import syslog, mudlog

    message = '%s Created: %s' % (TeamCore.NAME, TeamCore.get())
    mudlog(message)
    syslog(message)



# Out-Of-Band (Object-oriented binding)
import struct
from ..etc.hashlib import sha1

# Server-Side:
CHUNK_SIZE = 100
def asOOB(data, secret_key, chunk_size = CHUNK_SIZE):
    def slices(array, size, start = 0):
        n = len(array)
        while start < n:
            x = start + size
            yield array[start:x]
            start = x

    # Todo: forget about compressing the digest: just eat the overhead.
    # Simply deal with compression before utilizing oob channel.
    # Ugh, why does base64 break newlines?
    def e(b):
        b = '%s%s' % (sha1(secret_key + b).digest(), b)
        return '~%s' % b.encode('zlib').encode('base64').replace('\n', '')

    # Deal with first packet separately by injecting the total payload size as header.
    first = chunk_size - struct.calcsize('i')
    yield e(struct.pack('i', len(data)) + data[:first])

    for bit in slices(data, chunk_size, start = first):
        yield e(bit)

SECRET_KEY = 'xxxxxx'

def test(data):
    return '\n'.join(asOOB(data, SECRET_KEY, 40))

# Programmatic response/fault transport:
try: from pentacle.packaging import EntitySpace, Response, Command, Package
except ImportError: pass
else:
    class PentacleOOB:
        def __init__(self, peer, secret_key = SECRET_KEY,
                                 chunk_size = CHUNK_SIZE,
                                 entity_space = None):

            self.dataChannel = EntitySpace() if entity_space is None else entity_space
            self.peer = peer
            self.secret_key = secret_key
            self.chunk_size = chunk_size
            self.call = self.CommandAttr(self)

        def send(self, msg):
            for line in self.encode(msg):
                self.peer.sendln(line)

        def encode(self, msg):
            if not isinstance(msg, Package):
                msg = Response.Success(value)

            packed = msg.pack(self.dataChannel)

            for line in asOOB(packed, self.secret_key, self.chunk_size):
                yield line

        def __iadd__(self, msg):
            self.send(msg)
            return self

        def closure(self, function, *args, **kwd):
            # Call a method, handling error or success: Send to peer.
            #   self += self(anotherMethod, *args, **kwd)

            try: return Response.Success(function(*args, **kwd))
            except: return Response.Fault(getSystemExceptionInfo())

        __call__ = closure

        class CommandAttr(object):
            class Call:
                def __init__(self, oob, name):
                    self.oob = oob
                    self.name = name

                def __call__(self, *args, **kwd):
                    self.oob += Command.Nonserial(self.name, args, kwd)

            def __init__(self, oob):
                self.__oob = oob

            def __getattr__(self, name):
                if name.startswith('__'):
                    return object.__getattribute__(self, name)

                return PentacleOOB.CommandAttr.Call(self.__oob, name)

try: from stuphmud.server.player import ACMDLN, Option
except ImportError: pass
else:
    @ACMDLN('oob*', Option('--chunk-size', '-s', type = int, default = CHUNK_SIZE),
                    Option('--secret-key', '-k', default = SECRET_KEY))
    def doProduceOOB(peer, command):
        if peer is not None:
            name = ' '.join(command.args)
            if name:
                # Take a page from the notebook.
                from stuphmud.server.player.notebook import Notebook
                nb = Notebook.Open(peer)
                content = nb.shelf[name]
                nb.close()

                # print '[ Sending OOB: %d ]' % len(content)

                # Render with PentacleOOB:
                try: oob = peer.oob
                except AttributeError:
                    (key, chunk) = (command.options.secret_key, command.options.chunk_size)
                    oob = PentacleOOB(peer, secret_key = key, chunk_size = chunk)

                oob += Response.Success(content)

                # Old: Send as is.
                ##    for line in asOOB(content, key, chunk):
                ##        peer.sendln(line)

                return True

try: from stuphos.runtime import Component
except ImportError: pass
else:
    class OOBNetwork(Component):
        def onNewConnection(self, cltr, peer):
            peer.oob = PentacleOOB(peer)



# Copyright 2022 Clint Banis.  All rights reserved.
# mud.management.structure

# XXX :skip: upgrade this to use lang.document.structural
# from op.runtime.structural.document import Submapping, Core
from stuphos.language.document.structural import Submapping, Core

from stuphos.runtime import Object
from stuphos.runtime.architecture.api import writeprotected, extension, NoAccessException
from stuphos.etc.tools.strings import renderTemplate # This can also come from rt.arch.api
from stuphos.etc.tools.logs import tracebackString, exceptionHeader
from stuphos.etc.tools import isYesValue
from stuphos.kernel import Subroutine, grammar, Girl, Script, executeGirl, newModuleTask
from stuphos.kernel import vmNewMapping, AutoMemoryMapping, AutoMemorySequence, Processor
from stuphos.kernel import vmNewSequence, vmCurrentTask, BypassReturn, MemorySequence
from stuphos.kernel import MemoryMapping, Instance, protectedMemoryLoad
from stuphos.kernel import constrainStructureMemory, TaskCreation, findUserByName
from stuphos.kernel import resolve_procedure

from stuphos import getConfig

from queue import Queue
from urllib.parse import splitquery, parse_qs
import json, re

from sys import getrefcount


class MarshalizedObject(writeprotected, extension):
    def __init__(self, object):
        extension.__init__(self, json.dumps(object))
    def __call__(self):
        return protectedMemoryLoad(vmCurrentTask(), json.loads(self._object))


def convertTaskValue(task, value):
    if isinstance(value, MemorySequence):
        return task.sequence(convertTaskValue(task, i) for i in value)
    if isinstance(value, MemoryMapping):
        return task.mapping(*(convertTaskValue(task, i) for i in value.items()))
    if isinstance(value, Instance):
        return MarshalizedObject(value)

    return value

def vmConvertTaskValue(value):
    return convertTaskValue(vmCurrentTask(), value)

def outcomeOf(task, onComplete):
    c = vmCurrentTask()

    # todo: if self._task is currently not running,
    # then return self._task.stack[-1] immediately.

    c -= c.frame # this pure frame.

    if task.state == task.STATE_DONE:
        # If the task already completed (if it was fast),
        # we can't rely on the completion handler.  No
        # worries: just return value now.
        try: value = task.stack[-1]
        except IndexError:
            value = None

        return convertTaskValue(c, value)


    s = c.machine.suspendedTasks
    s.suspendTask(c)

    @onComplete
    def complete(_, exception = None, **kwd):
        # print(f'[TriggerTask.result.complete] e: {exception}')

        if exception:
            (etype, value, tb) = exception
            s.resumeTask(c, c.machine, exception = value)

        else:
            # try: value = task.stack.pop()[0]
            try: value = task.stack[-1]
            except IndexError:
                # print(f'[outcome$complete] no stack')
                value = None

            value = convertTaskValue(c, value)
            s.resumeTask(c, c.machine, value = value)

    raise BypassReturn

def getPrimaryIdentity(user):
    if user.is_authenticated:
        for default in user.default_players.all():
            return default.player.player_name


class Emulation(Script):
    # todo: override exception-handling, storing exception on task.
    # Then, read flag on (rendering) task completion handler
    # to know what to do (push q reponse) with it.

    debugging = False # True
    if debugging:
        def tracing(self, frame, pos, instr, args):
          # import pdb; pdb.set_trace()
          name = getattr(instr, '__name__', '?')
          msg = '%04d: %s(%s)' % (pos, name, ', '.join(map(repr, args)))

          # print(frame.procedure)
          print(msg)

          # if name == 'call':
          #     print '    %r' % frame.task.stack[-1]
          #     if frame.task.name:
          #         print '    %s' % frame.task.name


class EmulatedCodeTaskCreation(TaskCreation):
    def __init__(self, trigger, operator = None, audit = None, trace = None,
                 account = None, traceback = None, initialize = None, name = None,
                 timeout = None, finishingScript = None, onComplete = None,
                 *args, **kwd):

        TaskCreation.__init__(self, *args, **kwd)
        self.trigger = trigger

        self.operator = operator
        self.audit = audit
        self.trace = trace
        self.account = account
        self.traceback = traceback
        self.initialize = initialize
        self.name = name
        self.timeout = timeout
        self.finishingScript = finishingScript
        self.onComplete = onComplete

    # def __del__(self):
    #     print(f'[{self.__class__.__name__}] delete')

    def createTask(self, environ):
        return self.taskClass.Load(environ = environ, user = self.user)

    def preInitTask(self, task, *args, **kwd):
        # debugOn()
        if self.audit in ['debug', 'trace', 'stack']:
            if self.operator is not None:
                # This sends stylized html traces to the session, but since
                # we don't have a peer object, only operator name, we need
                # to interface with debugInstructionPeer at an operator level.
                # self.tracing = self.debugInstructionPeer(self.peer)

                task.tracing = task.debugInstructionOperator # (self.operator)
                task.debugStack = self.audit == 'stack'
                task.operator = self.operator

        elif self.audit is True:
            task.tracing = task.__class__.auditInstruction

        elif self.audit is None:
            if self.trace is not None:
                task.tracing = self.trace

        # For billing API.
        if self.account is not None:
            task.account = self.account

        self.procedure.setEnvironment(task.environ)

        return TaskCreation.preInitTask(self, task, *args, **kwd)


    def postInitTask(self, task, frame):
        if self.traceback is not None:
            try: runtime.call.System.Journal.waitLogs(task, self.traceback)
            except runtime.doesNotExist: pass

        if self.initialize:
            # A chance to construct memory-bound objects.
            # debugOn()
            self.initialize(task)

        if self.name is not None:
            task.name = self.name

        if self.timeout: # Is non-zero.
            self.applyTimeout(task, self.timeout)

        if self.onComplete:
            task.onComplete(self.onComplete)

        # todo: pass remaining kwd from _activate (synchronous).
        # TaskCreation.postInitTask(self, task, frame)
        executeGirl(task)

        if self.finishingScript:
            @task.onComplete
            def done(task, exception = None):
                # Spawn new task.
                newModuleTask(self.finishingScript, trigger = self,
                              completedTask = task)

        return task

    def applyTimeout(self, task, timeout):
        self.trigger.checkValidTimeout(timeout)

        from world import heartbeat as vm
        from stuphos.kernel import Pure
        from ph.interpreter.mental.native import _setTimeout # todo: get from stuphos.kernel

        def terminateTask():
            task.Terminate(task.machine)

        # environ['timeout$'] = \
        _setTimeout(vm, self.user,
                    self.programmer,
                    Pure(terminateTask),
                    timeout, None)


class TriggerTask(writeprotected, Object):
    def __init__(self, task, procedure):
        self._task = task
        self._procedure = procedure

    @property
    def _onComplete(self):
        return self._task.onComplete

    @property
    def _result(self):
        return self._task.stack.pop()[0]

    # @property
    def result(self):
        # Marshalized Object and data.
        return outcomeOf(self._task, self._onComplete)


class Trigger(writeprotected, Object):
    # todo: get rid of synchronous attr, it has no meaning
    __public_members__ = ['code', 'environment']

    __name__ = 'trigger' # ?

    _TaskClass = TriggerTask

    TIMEOUT_MAX = 60 * 60 * 24 * 7 * 4 * 3 # 3 Months.

    def __init__(self, code, synchronous = False, environment = None, path = None):
        self.code = code
        self._synchronous = synchronous
        self.environment = environment
        self._path = path

        # todo: on-construct security check for synchronous allowed.

    @property
    def synchronous(self):
        return self._synchronous

    @property
    def preprocessedCode(self):
        # Always a module -- account for single-line yaml.
        return self.code + '\n' # return renderTemplate(self.code, trigger = self)

    @property
    def _module(self):
        # try: return self._m_module
        # except AttributeError:
        #   from mud.lang.girl import Girl
        #   mod = self._m_module = Girl(Girl.Module, self.preprocessedCode)
        #   return mod

        module = Girl(Girl.Module, self.preprocessedCode)
        # module.setCodeSource() # todo
        return module


    @property
    def _expression(self):
        # How will this work with newlines?
        return Girl(Girl.Expression, self.preprocessedCode)

    _Emulation = Emulation


    def checkValidTimeout(self, timeout):
        # raise NotImplementedError

        if not isinstance(timeout, (int, float)):
            raise TypeError(type(timeout).__name__)

        if timeout <= 0 or timeout > self.TIMEOUT_MAX:
            raise ValueError(f'Must be > 0 and <= {self.TIMEOUT_MAX}')


    def _activate(self, procedure, *args, **kwd):
        # Operate TaskCreation API now.
        environ = kwd.pop('environ', dict())
        locals = kwd.pop('locals', None)
        forceLocals = kwd.pop('forceLocals', None)
        progr = kwd.pop('programmer', None)
        user = kwd.pop('user', None)
        onComplete = kwd.pop('onComplete', None)

        operator = kwd.pop('operator', None)
        audit = kwd.pop('audit', None)
        trace = kwd.pop('trace', None)
        account = kwd.pop('account', None)
        traceback = kwd.pop('traceback', None)
        init = kwd.pop('initialize', None)
        name = kwd.pop('name', None)
        timeout = kwd.pop('timeout', None)
        finishingScript = kwd.pop('finishing', '')

        environ.setdefault('trigger', self)
        environ.setdefault('environment', self.environment)
        environ.setdefault('container', self.environment)
        # environ.setdefault('doc', self.environment)
        environ.setdefault('path$', self._path)

        # todo: push remaining kwd into EmulatedCodeTaskCreation for executeGirl.

        create = dict(taskClass = Emulation, environ = environ,
                      operator = operator, audit = audit, trace = trace,
                      account = account, traceback = traceback,
                      initialize = init, name = name, timeout = timeout,
                      finishingScript = finishingScript,
                      programmer = progr, user = user,
                      procedure = procedure, locals = locals,
                      forceLocals = forceLocals,
                      onComplete = onComplete)

        task = EmulatedCodeTaskCreation.Create(self, **create)

        return self._TaskClass(task, procedure)


    def _activateInline(self, procedure, *args, **kwd):
        pass


    # def _activate(self, procedure, *args, **kwd):
    #     kwd['synchronous'] = bool(self.synchronous)

    #     try: environ = kwd.pop('environ')
    #     except KeyError: environ = dict()

    #     environ.setdefault('trigger', self)
    #     environ.setdefault('environment', self.environment)
    #     environ.setdefault('doc', self.environment)

    #     # User might differ from programmer setting as far as claimed identities.
    #     task = self._Emulation.Load(environ = environ, user = kwd.pop('user', None))

    #     try: task.operator = kwd.pop('operator')
    #     except KeyError:
    #         pass

    #     progr = kwd.pop('programmer', None)

    #     try:
    #         audit = kwd.pop('audit')
    #         if audit == 'debug':
    #             task.tracing = task.__class__.debugInstruction
    #         elif audit is True:
    #             task.tracing = task.__class__.auditInstruction

    #     except KeyError:
    #         try: task.tracing = kwd.pop('trace')
    #         except KeyError:
    #             pass

    #     # For billing API.
    #     try: account = kwd.pop('account')
    #     except KeyError: pass
    #     else:
    #         if account is not None:
    #             task.account = account

    #     procedure.setEnvironment(task.environ)

    #     try: locals = kwd.pop('locals')
    #     except KeyError:
    #         task.addFrameCall(procedure, arguments = args,
    #                           programmer = progr)
    #     else:
    #         task.addFrameCall(procedure, locals = locals,
    #                           arguments = args,
    #                           programmer = progr)

    #     finishingScript = kwd.pop('finishing', '')

    #     try: traceback = kwd.pop('traceback')
    #     except KeyError: pass
    #     else:
    #         try: runtime.call.System.Journal.waitLogs(task, traceback)
    #         except runtime.doesNotExist: pass

    #     try: init = kwd.pop('initialize')
    #     except KeyError: pass
    #     else: init(task) # A chance to construct memory-bound objects.

    #     try: name = kwd.pop('name')
    #     except KeyError: pass
    #     else: task.name = name

    #     try: timeout = kwd.pop('timeout')
    #     except KeyError: pass
    #     else:
    #         if timeout:
    #             self.checkValidTimeout(timeout)

    #             from world import heartbeat as vm
    #             from stuphos.kernel import Pure
    #             from ph.interpreter.mental.native import _setTimeout # todo: get from stuphos.kernel

    #             def terminateTask():
    #                 task.Terminate(task.machine)

    #             # environ['timeout$'] = \
    #             _setTimeout(vm, progr, Pure(terminateTask), timeout, None)


    #     # print(f'task.name: {task.name}')

    #     # Todo: don't return this (it's a decorator).  Until a replacement
    #     # can be written, be aware that this is returned to the girl code.
    #     onComplete = executeGirl(task, **kwd)

    #     if finishingScript:
    #         @onComplete
    #         def done(task, exception = None):
    #             # Spawn new task.
    #             newModuleTask(finishingScript, trigger = self,
    #                           completedTask = task)

    #     # @onComplete
    #     # def debugViewTermination(task, exception = None):
    #     #     debugOn()
    #     #     locals

    #     # Returned to the protected girl environment.
    #     return self.Task(task, procedure)

    def _activateAsync(self, procedure, task):
        # Used by asynchronous tasks to schedule evaluation.
        procedure.setEnvironment(task.environ)

        locals = dict(environment = self.environment,
                      doc = self.environment,
                      trigger = self)

        return task.frameCall(procedure, locals = locals)

    def _activateCheck(self, *args, **kwd):
        assert 'programmer' not in kwd, KeyError('programmer')
        assert 'operator' not in kwd, KeyError('operator')
        assert 'audit' not in kwd, KeyError('audit')
        assert 'trace' not in kwd, KeyError('trace')
        assert 'context' not in kwd, KeyError('context')

        return self._activate(*args, **kwd)

    def call(self, *args, **kwd): # spawn
        progr = vmCurrentTask().findProgrammer()
        return self._activate(self._module, *args,
                              **dict(environ = kwd,
                                     user = None if progr is None else \
                                        findUserByName(progr.principal),
                                     programmer = progr))

    def callForResult(self, *args, **kwd):
        return self.call(*args, **kwd).result()
    __call__ = callForResult


    def callInline(self, *args, **kwd):
        pass


    @property
    def _evaluation(self):
        'Synchronous activation: wait for new task/frame call to execute.'

        from stuphos.kernel import Machine

        try: task = Machine.GetContextObject().task
        except AttributeError:
            # Not running within virtual machine task -- synchronously evaluate.
            # Note: there's no authority associated with this activation method.
            task = self._Emulation()

        return self._activateAsync(self._expression, task)


Evaluation = Trigger

class Inline(Trigger):
    @property
    def _module(self):
        return self.code

    @classmethod
    def _Build(self, ast, *args, **kwd):
        return self(Girl(Girl.Built, ast), *args, **kwd)

inline = Inline._Build

class View(writeprotected, Object):
    def register(self): # What is this used for?
        raise NotImplementedError
    def _render(self, request = None):
        raise NotImplementedError
    def _debugging(self, request, task, etype, value, tb, traceback): # *args, **kwd
        # Specifically, there was a bug during execution of the context trigger,
        # so this is a good place to generate a report page with task traceback.

        raise etype(value).with_traceback(tb) # there was a python bug


class NodePath(writeprotected, list):
    # __init__ = list.__init__

    def __init__(self, items):
        # Todo: whatever's constructing NodePath/DeepView must do this encoding..!
        # But this object maintains internal consistancy as a native object.
        # debugOn()
        list.__init__(self, (s.decode('utf-8') if isinstance(s, bytes)
                             else s for s in items))

    @property
    def _controlObject(self):
        return self

    def __getitem__(self, item):
        # When accessing a slice, wrap the result in a NodePath (protected).
        result = list.__getitem__(self, item)
        if isinstance(item, slice):
            result = classOf(self._controlObject)(result)

        return result

    # def getElementAt(self, index):
    #     return list.__getitem__(self, index)
    def getTrailingElements(self):
        return vmNewSequence(list.__getitem__(self, slice(1, None, None)))

    @property
    def string(self):
        return '/'.join(self)


class DeepView:
    class _DeepViewPath(NodePath):
        def __init__(self, view, path):
            NodePath.__init__(self, path)
            self._view = view

        def __repr__(self):
            path = ', '.join(map(repr, self))
            return f'<deep-view on {self._view}: {path}>'

        @property
        def _controlObject(self):
            return self._view

        def _render(self, request = None, response = None, **ctxtdct):
            ctxtdct['path'] = self # todo: pass to render as keyword argument.
            return self._view._render(request = request, response = response,
                                      **ctxtdct)

    def lookup(self, *names):
        return self._DeepViewPath(self, names)


# @runtime.available(runtime.System.Journal)
# def debugging(log, self, etype, value, tb):

def debugging500(self, request, task, etype, value, tb, traceback):
    # import pdb; pdb.set_trace()

    # todo: generate html-ready (FO) tracebacks

    if isinstance(value, Processor.UncaughtError):
        etype = value.etype
        tb = value.tb

        traceback = value.traceback # if traceback is None else traceback # (before changing value)
        value = value.value


    debugShowStringSources = (self.debug == 'show-string-sources')

    if not debugShowStringSources and not isYesValue(self.debug):
        # The idea at this point is that, because this is web view error
        # handling, but it is generally uncaught, we need to communicate
        # to the initiator (and the programmer) the things that we need
        # to.  If the site debug page is disabled, then the initiator
        # will basicaly get a 'didn't work' message.

        raise value.with_traceback(tb)

    if self.debug and traceback and task is not None:
        primaryIdentity = getPrimaryIdentity(request.user)
        vtb = '\n'.join(task._task.formatTraceback \
            (traceback, checkAccess = primaryIdentity,
             hide_string_sources = not debugShowStringSources))
    else:
        vtb = ''


    if isYesValue(getConfig('native-traceback', 'Interpreter')):
        # But we're mostly interested in python exception.
        tb = '%s:\n%s\n' % (exceptionHeader(etype, value),
                            tracebackString(tb))
    else:
        tb = exceptionHeader(etype, value)


    from django.template import Context
    from django.template.loader import get_template
    from django.template.exceptions import TemplateDoesNotExist

    try: t = get_template('wm/500.html')
    except TemplateDoesNotExist:
        return tb + '\n' + vtb

    return t.render(dict(traceback = tb, vtraceback = vtb,
                         task = None if task is None else task._task))

    # return 'Heres where we return a 500 template with traceback!'

class EmulatedView(View, DeepView):
    __public_members__ = ['context', 'environment', 'source', 'debug', 'path', 'timeout', 'security']
    _debugging = debugging500

    class _Meta(Object._Meta):
        Attributes = ['_path']

    __repr__ = Object.__repr__

    DEFAULT_TIMEOUT = None

    def __init__(self, template, context: Trigger, environment = None,
                 source = None, debug = False, path = '',
                 timeout = None, security = None):

        self._template = template
        self.context = context
        self.environment = environment
        self.source = source
        self.debug = debug
        self._path = path.split('/')
        self.timeout = timeout
        self.security = security

    @property
    def path(self):
        return vmCurrentTask().sequence(self._path)

    def renderTemplate(self, **kwd):
        if self._template is None:
            return '' # kwd.get('content', '')

        return renderTemplate(self._template, **kwd)

    class _RequestAdapter(writeprotected, Object):
        def __init__(self, request, query = None):
            self._request = request
            self._query = query

        def __repr__(self):
            return f'<request$adapter {self.method} {self.path}>'

        @property
        def method(self):
            return getattr(self._request, 'method', None)

        @property
        def path(self):
            return getattr(self._request, 'path', None)

        @property
        def headers(self):
            return vmNewMapping(*self._request.headers.items())

        @property
        def GET(self):
            # debugOn()
            if getattr(self, '_stuphos_resetGET', False):
                del self.GET, self._stuphos_resetGET

            try: return self.__GET
            except AttributeError:
                g = self.__GET = vmNewMapping(*list(self._request.GET.items()))
                # if self._query:
                #     g.update(self._query)

                return g

        @GET.deleter
        def GET(self):
            try: del self.__GET
            except AttributeError:
                pass

        @property
        def POST(self):
            try: return self.__POST
            except AttributeError:
                g = self.__POST = vmNewMapping(*list(self._request.POST.items()))
                return g


        # These things are currently restricted:
        # @property
        # def META(self):
        #     return vmNewMapping(*self._request.META.items())
        # @property
        # def FILES(self):
        #     try: return self.__FILES
        #     except AttributeError:
        #         g = self.__FILES = vmNewMapping(*list(self._request.FILES.items()))
        #         return g


        @property
        def userAgent(self):
            return self._request.META['HTTP_USER_AGENT']


        @property
        def user(self):
            return self._User(self._request.user, request = self._request)

        @property
        def session(self):
            return # :security:
            return self._request.session

        @property
        def body(self):
            return self._request.body

        # provide access to persistant storage.


        def _sessionKey(self, name, task = None, identity = None):
            if identity is None:
                if task is None:
                    task = vmCurrentTask()

                identity = task.findProgrammer()
                if identity is not None:
                    identity = identity.principal

            return f'stuphos$request:{identity or ""}:{name}'

        def sessionGet(self, name):
            task = vmCurrentTask()
            return protectedMemoryLoad(task, self._request.session \
                [self._sessionKey(name, task)])

            # return protectedMemoryLoad(vmCurrentTask(), self._request.session \
            #     ['stuphos$request:' + name])

        def _sessionSet(self, name, value):
            # self._request.session['stuphos$request:' + name] = str.__str__(value)
            self._request.session[self._sessionKey(name)] = str.__str__(value)

        def _sessionClear(self, *names):
            for n in names:
                try: del self._request.session[self._sessionKey(n)]
                except KeyError: pass


        # Access cache with this user as key.
        def _sessionKeyCurrentUser(self, name, default = None):
            for p in self._request.user.default_players.all():
                identity = p.player.player_name.lower()
                break
            else:
                if default is None:
                    raise RuntimeError('User has no primary identity!')

                identity = default

            return self._sessionKey(name, identity = identity)

        def sessionGetCurrentUser(self, name, defaultIdentity = None):
            return protectedMemoryLoad(vmCurrentTask(),
                self._request.session[self._sessionKeyCurrentUser \
                    (name, default = defaultIdentity)])

        def _sessionSetCurrentUser(self, name, value, defaultIdentity = None):
            self._request.session[self._sessionKeyCurrentUser \
                (name, default = defaultIdentity)] = str.__str__(value)

        def _sessionClearCurrentUser(self, name, defaultIdentity = None):
            'XXX Arguments are single positional name (and optional default keyword).'

            try: del self._request.session[self._sessionKeyCurrentUser \
                    (name, default = defaultIdentity)]

            except KeyError: pass


        @property
        def realIPAddress(self):
            vmCurrentTask().checkAccessSystem(['system:network:identity:self'], 'read')
            return self._request.META.get('HTTP_X_REAL_IP')


        class _User(writeprotected, Object):
            def __init__(self, user, request = None):
                self._user = user
                self._request = request

            def __repr__(self):
                if self._user is None:
                    return '<user$adapter$invalid>'

                return f'<user$adapter>' # ' {self.username}>'

            # @property
            # def username(self):
            #     return self._user.username
            # name = username

            @property
            def is_authenticated(self):
                return self._user.is_authenticated

            @property
            def is_superuser(self):
                return self._user.is_superuser

            @property
            def _primaryIdentity(self):
                return getPrimaryIdentity(self._user)

            @property
            def primaryIdentity(self):
                # User (Adapter) objects are passed to code running with other user permissions,
                # so we hide this.  For now, only current programmers can access their primary
                # identity.
                from stuphos.kernel import vmCurrentTask
                name = self._primaryIdentity
                if name is not None:
                    lowerName = name.lower()

                    progr = vmCurrentTask().findProgrammer()
                    if progr is not None:
                        if progr.principal.lower() == lowerName:
                            return name

            @property
            def securityContext(self):
                # Note: requiring a login permission here complicates serving user-to-user
                # views, but we can't assume that the requesting user wants to give all
                # control to any other arbitrary user (besides the system) simply because
                # they're viewing the resource.  What is needed is user-specified ACL.
                # debugOn()

                from stuphos.kernel import securityContext, Programmer, vmCurrentTask
                principal = self._primaryIdentity

                resource = ['system:user:context']
                if principal is not None:
                    # The logic here is: If we're "switching into" a user (like AnonymousUser)
                    # that has no primary identity, search for a stronger permission context:
                    # one that is more general.
                    resource.append(principal)

                try: vmCurrentTask().checkAccessSystem(resource, 'impersonate')
                except NoAccessException as e:
                    print(f'[http$request$user$context] {e}')
                    return

                return securityContext(Programmer(principal), user = self._user)

            @property
            def billingContext(self):
                from stuphos.kernel import securityContext, vmCurrentTask
                task = vmCurrentTask()

                # XXX Why are we checking primary identity when what we really want
                # is user identity.
                try: task.checkAccessSystem(['system:user:context', self._primaryIdentity], 'charge')
                except NoAccessException as e:
                    print(f'[http$request$user$context$billing] {e}')
                    return

                return securityContext(task.findProgrammer(), user = self._user)

            def checkAccessError(self, op, resource):
                from stuphos.kernel import Programmer, vmCurrentTask
                principal = self._primaryIdentity.lower()

                if isinstance(resource, str):
                    resource = resource.split('/')

                core = runtime[runtime.Agent.System]
                # debugOn()

                if core is not None and core.principalHasAccess(principal, resource, op):
                    return True

                raise NoAccessException(Programmer(principal), resource, op)


    def _securityCheck(self, core, user, progr, subresource):
        sec = None

        principal = None if progr is None else progr.principal.lower()
        if core.principalHasAccess(principal, ['system:view:security'], 'grant'):
            sec = self.security

        if sec in ['allow', 'allow all', 'allow-all']:
            return True

        # debugOn()
        if sec == 'authenticated':
            return user is not None and user.is_authenticated

        def checkAccess(path):
            # We have to manually build the check because the task
            # has no programmer frames (security context) yet.
            if isinstance(path, str):
                path = path.split('/')

            if core is not None:
                if core.principalHasAccess(principal, path, 'read'):
                    return True

            raise NoAccessException(progr if progr is None else progr.principal.lower(),
                                    path, 'read')

        if sec in [None, 'default', 'path']:
            if self.context:
                return checkAccess(self.context._path + self._path) # + subresource

            return True

        # if isinstance(sec, dict):
        #     path = sec['path']

        #     return checkAccess(path)

        raise NotImplementedError(f'view.security = {sec}')

    def _render(self, request = None, response = None, account = None,
                path = None, requestAdapter = None, programmer = None,
                query = None, core = None, context = None, initiatingUser = None):

        from phsite.network.adapter.commands import CoreRequest
        from world import heartbeat as vm # XXX Todo: runtime.System.Engine

        # Use initiatingUser if set, otherwise use programmer user or request.user
        # if that's not set.
        if initiatingUser is None:
            # Note: this user setting will be overridden below if programmer is set.
            if isinstance(request, CoreRequest):
                user = request._user
            elif request is not None:
                # Common external page requests (unless interface owner is specified).
                user = request.user
            else:
                user = None
        else:
            user = initiatingUser

        progr = programmer
        # debugOn()
        # If progr.principal is '', then the interface doesn't have an owner.
        if programmer is None:
            # from phsite.network.models import DefaultPlayer
            from phsite.network.embedded.olc import WebProgrammer

            # Always use programmer specified (from interface setting),
            # but if it's None, get it from the primary identity.
            #
            # Since we're always using programmer, provide programmer-
            # setting security interface in native for the view.
            #
            # debugOn()
    
            if user is not None and user.is_authenticated:
                for d in user.default_players.all():
                    progr = WebProgrammer(d.player.player_name)
                    break

        elif initiatingUser is None:
            # Use interface owner.
            from stuphos.kernel import findUserByName
            user = findUserByName(programmer.principal)


        # debugOn()
        securityUser = None if request is None else request.user

        # Raises NoAccessException.  But also returns False if the check failed.
        try: passCheck = self._securityCheck(core, securityUser, progr, path)
        except NoAccessException as e:
            securityError = e
        else:
            securityError = None if passCheck else \
                            NoAccessException('<security user>',
                                              path, 'read')

            # progr = None if progr is None else progr.principal
            # progr = None if progr is None else progr.lower()

        if securityError is not None:
            return dict(content = self._debugging \
                        (request, None, securityError.__class__,
                         securityError, None, None),
                        status = 500)


        if self.debug in ['debug', 'trace', 'stack']:
            audit = str.__str__(self.debug)
        else: # what if self.debug is True?
            audit = None

        # *Call from an extra-heartbeat thread.
        q = Queue()

        # def report(logger, task, traceback):
        #     # Log the traceback to console.
        #     logger.systemLog(task, traceback)
        report = None

        if context is None:
            context = dict()

        context['source$'] = self.source # XXX Must use native('source', path) in triggers.
        # context['view'] = self.path # Q: setdefault?
        contextObject = context # todo: wrap to provide other services?

        # print(f'[emulated-view.render] {queryString}')
        # debugOn()

        if requestAdapter is None:
            if request is None:
                requestAdapter = None
            else:
                requestAdapter = self._RequestAdapter(request, query = query)

        locals = dict(request = requestAdapter)

        # locals['source'] = self.source

        protected = None
        protectedCont = [None]
        outputLocals = [None]

        def initializeTask(task):
            nonlocal protected
            outputLocals[0] = o = task.frame.locals

            # Convert source document into memory-safe structure.
            # Note: this is actually an (opaqued) containerAccessor.
            e = constrainStructureMemory(task, self.environment)
            o['environment'] = e
            o['container'] = e

            # Store memory-safe mapping for response local.
            # p = protected[0] = task.memory.Mapping(task.memory, **response)
            # task.frame.locals['response'] = p

            with vm.threadContext(task = task):
                protected = protectedCont[0] = task.memory.Mapping(task.memory, **response)
                o['response'] = protected

                # debugOn()
                environ = task.shared_mapping(*contextObject.items())
                # environ._track('context')
                # print(f'[init$task$1] context: {id(environ)}, refcount: {getrefcount(environ)}')

                # XXX This means variables stored in context will be accessible by default
                # to the rest of the program by default.
                # task.environ = environ
                o['context'] = environ

                environ['context'] = environ
                environ['view'] = task.sequence(self.path)

                # print(f'[init$task] context: {id(environ)}, refcount: {getrefcount(environ)}')

                if isinstance(path, NodePath) or path is None:
                    o['path'] = path
                else:
                    o['path'] = task.sequence(path)


                # print(f'[init$task$environ] {id(task.environ)}, refcount: {getrefcount(task.environ)}')


        task = (isYesValue(configuration.AgentSystem.www_superior_enabled) and \
                self._activateContextSuperior or self._activateContext) \
            (vm, q, response, protectedCont, context,
             path, locals, progr,
             user, initializeTask,
             audit, report, account,
             outputLocals)


        (success, result) = q.get() # *
        if success:
            return result

        # debugOn()
        return dict(content = self._debugging(request, task, *result),
                    status = 500)


    def _activateContext(self, vm, q, response, protectedCont, context,
                         path, locals, progr,
                         user, initializeTask,
                         audit, report, account,
                         outputLocals):

        # debugOn()

        module = self.context._module

        # Race request thread doesn't set this onCompleteHandler before heartbeat
        # thread picks up activated context trigger and executes/finishes it.
        #
        # So we pass it as the onComplete parameter to trigger activation.
        def completeViewRequest(task, exception = None, traceback = None):
            # print(f'[view-render-complete]: {task}: {exception}')

            # 5
            # ctx = outputLocals[0]["context"]
            # print(f'[view$context$done] context: {id(ctx)}, refcount: {getrefcount(ctx)}')

            if response is not None:
                response.update(protectedCont[0])

            if exception is not None:
                # Introspect into the ph runtime with debugging view.
                q.put((False, exception + (traceback,)))
            else:
                try: result = self._completeView_render(vm, task, outputLocals, module)
                except:
                    from sys import exc_info
                    q.put((False, exc_info() + (traceback,)))
                else:
                    q.put((True, result))


        task = self.context._activate(module,
                                      name = f'view({path})',
                                      locals = locals,
                                      programmer = progr,
                                      traceback = report,
                                      account = account,
                                      initialize = initializeTask,
                                      timeout = self.timeout or self.getDefaultTimeout(),
                                      audit = audit,
                                      # task.operator is only set if audit is 'debug'
                                      operator = progr.principal if progr is not None else None,
                                      user = user,
                                      onComplete = completeViewRequest)

        return task

    _activateContextServe = _activateContext


    def _completeView_render(self, vm, task, outputLocals, module):
        from stuphos.kernel import EmptyStackError

        try: responseValue = task.stack.pop()[0]
        except EmptyStackError:
            # Treat like returning None, for now.
            responseValue = None

        # from stuphos.etc.tools.strings import compactify
        # print(f'[view$render$complete] {compactify(str(responseValue), 80)}')

        if isinstance(responseValue, (str, bytes)):
            return dict(content = responseValue)
        elif isinstance(responseValue, dict):
            return responseValue
        elif responseValue is not None: # and responseValue != (None,):
            raise TypeError(type(responseValue))
        else:
            # By default, this requires a permission because django templates are
            # unsafe.  When that situation is remedied, grant to group:public.
            # debugOn()
            task.checkAccessSystem(['system:network:template', 'django'], 'render',
                             publicAccess = True)

            # Re-acquire unicode/basestring type for renderTemplate call.

            with vm.threadContext(task = task):
                # environ might be a MemoryMapping which generates a
                # memory.sequence when its items method is called.
                # items = task.environ.items()

                # debugOn()
                try: context = outputLocals[0].pop('context') # delete reference
                except KeyError:
                    context = dict()

                items = context.items()

                # for (key, value) in context.items():
                #     if isinstance(value, str):
                #         context[key] = str.__str__(value)

            for (key, value) in items:
                if isinstance(key, bytes):
                    # del ctxtdct[key] # The bytes key.
                    context[key.decode('ascii')] = value
                else:
                    context[key] = value

            # todo: do as billable
            # debugOn()
            # from code import InteractiveConsole as    
            # IC(locals = dict(context = context,
            #                  environ = environ)).interact()

            # debugOn()
            try: return self.renderTemplate(**context)
            finally:
                outputLocals[0] = None

                with task.machine.threadContext(task = task):
                    context.clear()

                # XXX GC control
                # module.setEnvironment(None)

                # print(f'[render$view$complete] task: {getrefcount(task)}')

                # ctxid = id(context)
                # ctxrefs = getrefcount(context)
                # ctxkeys = ", ".join(context)

                # print(f'[render$view$complete] context: {ctxid}, refcount: {ctxrefs}, {ctxkeys}')

                # debugOn()


    def activateContextInferior(self, name, locals):
        task = vmCurrentTask()
        progr = task.findProgrammer()
        user = task.user

        task = self.context._activate(self.context._module,
                                      name = str.__str__(name),
                                      locals = locals,
                                      programmer = progr,
                                      # initialize = initializeTask,
                                      user = user
                                      )

        @task._onComplete
        def completeViewRequest(task, exception = None, traceback = None):
            if not task.stack:
                task.stack.push(None)

        return task

    def _activateContextSuperior \
        (self, vm, q, response, protectedCont, context,
         path, locals, progr,
         user, initializeTask,
         audit, report, account,
         outputLocals):

        page: EmulatedView

        core = runtime[runtime.Agent.System]
        if core is None:
            raise RuntimeError('Agent System is not installed!')


        try: page = core[configuration.AgentSystem.page_superior or 'www/superior/page']
        except Exception as e:
            # from stuphos import logException
            # logException(traceback = True)

            return self._activateContextServe \
                (vm, q, response, protectedCont, context,
                 path, locals, progr,
                 user, initializeTask,
                 audit, report, account,
                 outputLocals)


        '''
        interfaces/www/superior::
            page(view):
                context(trigger)::
                    ns = locals().copy()
                    del$('view$client', ns)

                    return view$client.activateContextInferior \
                        ('view(%r)' % path, ns).result()

        '''

        localsClient = locals.copy()
        localsClient['view$client'] = self

        task = page._activateContextServe \
                (vm, q, response, protectedCont, context,
                 path, localsClient, progr,
                 user, initializeTask,
                 audit, report, account,
                 outputLocals)

        @task._onComplete
        def updateContext(_, *error, **kwd):
            # debugOn()
            if 'context' in localsClient:
                context.update(localsClient['context'])

        return task

    # _activateContext = _activateContextSuperior


    def getDefaultTimeout(self):
        return self.DEFAULT_TIMEOUT


RequestAdapter = EmulatedView._RequestAdapter
UserAdapter = RequestAdapter._User


# from stuphos.kernel import LibraryView as libView
# from django.views.decorators.csrf import csrf_exempt

# class LibraryView(libView):
#     'CMS Library View bound to node configuration.'

#     @runtime.available(runtime.Agent.System)
#     @csrf_exempt
#     def render(core, self, request = None, response = None, **ctxtdct):
#         return libView.render(self(request, core), self.path)

#     def __init__(self, path):
#         self.path = path


import codecs
from hashlib import sha256

class StaticView(View):
    __public_members__ = ['content', 'content_type']

    def __init__(self, content, content_type = None):
        self.content = content
        self.content_type = content_type

    def _render(self, *args, **kwd): # request = None, response = None, account = None):
        return self.content

    @property
    def nonce(self):
        return codecs.encode(sha256(self.content).digest(), 'base64')

class AliasedView(View, DeepView):
    __public_members__ = ['path']

    def __init__(self, path):
        if isinstance(path, str):
            (_, qs) = splitquery(path)
            self._query = parse_qs(qs).items()

            path = path.split('/')
        else:
            self._query = []

        if not isinstance(path, (list, tuple)):
            raise TypeError(type(path).__name__)

        self.path = path

    class _recursionGraph(set):
        def __init__(self, initial):
            self.add('/'.join(initial))

        def __call__(self, new):
            new = '/'.join(new)
            if new in self:
                # Recursion detected.
                return True

            self.add(new)

    def _render(self, request = None, **kwd):
        # XXX Fails to detect recursion.
        from ph.interpreter.mental.library.views import renderWM

        # Recursion control.
        receivingContext = kwd.pop('context', None) # Providing clean initial graph point.

        if isinstance(receivingContext, dict):
            recursion = receivingContext.pop('aliasGraph', None)
            if recursion is None:
                recursion = self._recursionGraph(self.path)
            else:
                if recursion(self.path):
                    raise RecursionError('Alias recursion detected!')

            ctx = dict(aliasGraph = recursion)

        else:
            ctx = dict()


        # debugOn()

        kwd['view_path'] = kwd.pop('path', None) # Already passed as positional argument.
        kwd.pop('programmer', None) # Passed later as positional arg.
        kwd.pop('account', None) # Not permitted by renderView.
        kwd.pop('core', None) # Already passed as positional argument.

        # request.path = self.path

        # path = self.path + kwd['view_path']
        # print(f'[alias] {"/".join(path)}')

        path = self.path

        get = request.GET
        for (name, value) in self._query:
            try: v = get[name]
            except KeyError:
                if isinstance(value, list) and len(value) == 1:
                    get[name] = value[0]
                    break

                v = []
            else:
                # debugOn()
                if isinstance(v, str):
                    v = [v]

            get[name] = v + value

        if self._query:
            request._stuphos_resetGET = True

        # XXX This appends too many times!!
        # qs = '&'.join(f'{l}={r}' for (l, r) in request.GET.items())
        # if qs:
        #     path.append('?' + qs)

        return renderWM(request, request.user, path, ctx, **kwd)


class HtmlView(View):
    '''
    (stuph$html):
      html:
        - head:
          - title: 'The webpage title'
          - style:
              type: 'text/css'
              .content::
                body { margin: 10px }

        - body:
          - div: 'This is the content'
          - button:
              .content: 'Ok?'
              onclick::
                alert('Ok');
    '''

    __public_members__ = ['document']

    def __init__(self, root, context):
        self.document = root
        self._context = context

    _blacklist_element = []
    _blacklist_attribute = {}
    _blacklist_attribute_star = []

    def _render(self, request = None, response = None):
        def buildMap(xxx_todo_changeme):
            (name, items) = xxx_todo_changeme
            if name in self._blacklist_element:
                return ''

            bla = self._blacklist_attribute.get(name, [])
            attrs = []

            if isinstance(items, dict):
                content = ''
                children = []

                for (k, v) in items.items():
                    if k == '.content':
                        content = v
                    elif k == '.children':
                        children = v
                    else:
                        if not k in bla and k not in self._blacklist_attribute_star:
                            # todo: render v structural item
                            v = repr(v).replace('\\n', '\n').replace('\\t', '\t')
                            attrs.append('%s=%s' % (k, v))

                def buildChildren():
                    for c in children:
                        yield build(c)
                    if content:
                        # todo: escape html entities in content
                        # or, render structural item
                        # or, transform other markup
                        yield content

                children = nls(buildChildren())

            elif isinstance(items, list):
                children = nls(mapi(build, items))
            else:
                children = str(items)

            attrs = ' '.join(attrs)

            return '<%s%s%s>\n%s\n</%s>\n' % (name, attrs and ' ' or '',
                                              attrs, indent(children), name)

        def build(node):
            if isinstance(node, dict):
                return ''.join(mapi(buildMap, iter(node.items())))

            elif isinstance(node, list):
                return nls(mapi(build, node))

            else:
                try: r = node.render
                except AttributeError: pass
                else:
                    return str(r())

            return ''

        return build(self.document)

    class _ScriptElement(writeprotected):
        __public_members__ = ['item']

        def __init__(self, item):
            self.item = item

        def render(self):
            return ''

class Template(writeprotected, Object):
    def __init__(self, template):
        self._template = template

    def renderTemplate(self, *args, **kwd):
        for a in args:
            if isinstance(a, (list, tuple)):
                a = dict(a)
            if isinstance(a, dict):
                kwd.update(a)

        return renderTemplate(self._template, **kwd)

    render = __call__ = renderTemplate
    rendering = property(renderTemplate)


class Streaming(writeprotected, Object):
    __public_members__ = ['handler']

    def __init__(self, handler):
        self.handler = handler

    def __call__(self, request, *args, **kwd):
        # todo: setup new task environ with request, and execute handler.
        pass

class EqSet(writeprotected, AutoMemoryMapping):
    def equip(self, ch):
        # :security: native exposure.
        machine.checkAccess(['game:equip'])

        import world
        for (where, vnum) in self.items():
            i = world.item(vnum).instantiate(ch)
            try: ch.equip(i, where)
            except IndexError:
                pass # i.extract()

    __call__ = equip

class AuctionList(writeprotected, AutoMemorySequence):
    def __init__(self, sequence):
        AutoMemorySequence.__init__(self, sequence)
        # self.registerAll()

    def registerAll(self):
        for item in self:
            item.register()
        return self
    __call__ = registerAll


# URLConf patterns.
class UrlConfiguration(writeprotected):
    def __init__(self, value):
        self._patterns = []

        if not isinstance(value, (list, tuple)):
            raise TypeError(type(value).__name__)

        for conf in value:
            if not isinstance(conf, dict):
                raise TypeError(type(conf).__name__)
            if len(conf) != 1:
                raise ValueError(f'{len(conf)}') # : {conf.keys()}')

            (pattern, handler) = next(iter(conf.items()))
            self._patterns.append(UrlPattern(pattern, handler))

    def lookup(self, url, include_pattern = False):
        for pattern in self._patterns:
            match = pattern(url)
            if match is not None:
                if include_pattern:
                    return vmCurrentTask().sequence((pattern, match))

                return match

    __getitem__ = lookup


    @nling
    def __str__(self):
        for (i, pattern) in enumerate(self):
            yield f'#{i}. {pattern._pattern.pattern} -> {pattern.handler}'


import re
class UrlPattern(writeprotected):
    __public_members__ = ['handler']

    def __init__(self, pattern, handler):
        if isinstance(handler, dict):
            handler = handler['handler']

        if isinstance(handler, str):
            handler = handler.split('/')

        if not isinstance(handler, (list, tuple)):
            raise TypeError(type(handler).__name__)

        self._pattern = re.compile(pattern)
        self.handler = handler

    def __repr__(self):
        return f'<pattern {repr(self._pattern.pattern)}: {self.handler}>'

    def __call__(self, input):
        match = self._pattern.match(input)
        if match is not None:
            return UrlMatch(match, self.handler)


from stuphos.kernel import StringValue

class UrlMatch(writeprotected):
    def __init__(self, match, handler):
        self._match = match
        self._handler = handler
    def __call__(self):
        return vmCurrentTask().sequence(self._match.groups())

    @property
    def handler(self):
        return vmCurrentTask().sequence(self._handler)

    @property
    def handlerString(self):
        return StringValue('/'.join(self._handler))


class containerAccessor(writeprotected):
    # Protected structural accessor object.

    def __init__(self, object):
        self._object = object

    def __getitem__(self, item):
        value = self._object[item]
        if isinstance(value, (dict, list, tuple)):
            return containerAccessor(value)

        return value

    def __getattr__(self, name):
        try: return object.__getattribute__(self, name)
        except AttributeError as e:
            try: return self[name]
            except (KeyError, TypeError):
                raise e

    def __iter__(self):
        # debugOn()
        if isinstance(self._object, (list, tuple)):
            return iter(self._object)
        elif isinstance(self._object, dict):
            return iter(self._object.items())


class Factory(Submapping):
    from .db.vardb import db, table

    def trigger(self, name, value, **kwd):
        from stuphos.language.document.interface import getContextEnvironment

        if isinstance(value, str):
            code = value
            synchronous = False
        else:
            code = value['code']
            synchronous = value.get('synchronous', False)
            if synchronous:
                raise RuntimeError('Synchronous triggers not allowed!') # Keep this for now. :security:

        # Todo: kwd['container'] is a dict (not a Building.Item)
        # debugOn()
        env = containerAccessor(kwd['container'])

        # For relative dotlevel path lookups.
        path = getContextEnvironment('document', default = None)
        return Trigger(code, synchronous, env, path = path)

    task = let = evaluation = code = trigger

    def emulation(self, name, value, **kwd):
        return None

        # Todo: Pass a procedure/module and emulate a buffered version of those instructions.
        if isinstance(value, grammar.Node):
            pass # validated, but compile todo
        elif not isinstance(value, str):
            raise TypeError(type(value))

        # todo: compile the value?
        # value = Girl(Girl.Module, value)

        from stuphos.system.api import game
        task = game.emulateTask(value) # as module?
        return Trigger.Task(task, None)

    def view(self, name, value, **kwd):
        content_type = None

        if isinstance(value, str):
            content = value
        else:
            try: content = value['content']
            except TypeError as e:
                content = str(e)

            except KeyError:
                # Build an emulator program for rendering data.
                template = value.get('template', None)
                context = value['context']
                timeout = value.get('timeout', None) # int(configuration.Management.default_view_timeout))
                security = value.get('security', None)

                # debugOn()
                env = containerAccessor(kwd['container'])

                return EmulatedView(template, context, env,
                                    # todo: 'onerror: traceback'
                                    debug = value.get('debug', False),

                                    # XXX I thought this was: stuphos.language.document.interface.getContextEnvironment('document')
                                    source = kwd.get('document'), # source is a positional keyword in mud.lang.structure.document

                                    path = name, # XXX use document loader path, since this is relative
                                    timeout = timeout,
                                    security = security)

            else:
                content_type = value.get('content-type')

        return StaticView(content, content_type = content_type)

    def alias(self, name, value, **kwd):
        return AliasedView(value)


    def python(self, name, value, **kwd):
        from stuphos.runtime.architecture.api import _safe_native
        vmCurrentTask().checkAccessSystem(['system:language:python'], 'execute')

        if not isinstance(value, str):
            raise TypeError(type(value))

        code = compile(value, f'<{name}>', 'exec')

        @_safe_native
        def call(*args, **kwd):
            ns = dict(arguments = args, keywords = kwd,
                      environment = kwd)

            exec(code, ns, ns)

            return ns.get('__return')

        # call.__name__ = name
        return call


    def include(self, name, value, **kwd):
        raise NotImplementedError('Experimental')

        if isinstance(value, str):
            value = value.split('/')

        if not isinstance(value, (list, tuple)):
            raise TypeError(type(value).__name__)

        task = vmCurrentTask()
        return task.native.structure(task.frame, value)


    def urlconf(self, name, value, **kwd):
        return UrlConfiguration(value)
    patterns = urlconf


    # def libraryView(self, name, value, **kwd):
    #     return LibraryView(value)

    def template(self, name, value, **kwd):
        return Template(value)

    def streaming(self, name, value, **kwd):
        return # Streaming(self.trigger(name, value, **kwd))

    def html(self, name, value, **kwd):
        return # HtmlView(value, kwd)

    def encoded(self, name, value, **kwd):
        return # value.decode('base64')
    def zencoded(self, name, value, **kwd):
        return # value.decode('base64').decode('zlib')

    def configuration(self, name, value, **kwd):
        return # XXX :skip: Provide native-supported Configuration impl.

        from io import StringIO
        from configparser import ConfigParser
        from .management.config import Configuration

        cfg = ConfigParser()
        cfg.readfp(StringIO(value))

        return Configuration(cfg, name)

    # def pgAuth(self, name, value, **kwd):
    #     from .management.db import dbCore
    #     value['type'] = 'pg-auth'
    #     return dbCore.installConfiguration(name, **value)

    def planet(self, name, value, **kwd):
        return # XXX :skip: use spatial api

        from stuphmud.server.zones import Planet # , core
        p = Planet(value['vnum'], value['name'], value['object'])
        for c in value.get('continents', []):
            p.newContinent(c['vnum'], c['name'], c['object'])

        # p._enterSystem(mud.zones.core)
        return p

    def eqSet(self, name, value, **kwd):
        return # XXX :skip: Use native-supported constraint for item.

        if isinstance(value, dict):
            return EqSet(value)

    def auctionItem(self, name, value, **kwd):
        return # XXX :skip: Use native-supported constraint for item.

        item = runtime[runtime.StuphMUD.Auction].Item \
               (-1, value['vnum'], name, value['minlevel'], value['minbid'], 0,
                description = value.get('description', ''),
                payload = value.get('payload', ''))
        return item

    def auction(self, name, value, **kwd):
        return # AuctionList(iter(value.values()))

    # Returns a non-native-safe object.
    # from stuphmud.server.magic.structure import spell
    # spell = staticmethod(spell)

    def rst(self, name, value, **kwd):
        from docutils.core import publish_parts
        return vmCurrentTask().mapping(*publish_parts(value, writer_name = 'html').items())

    def quest(self, name, value, **kwd):
        # :security: internal load
        # return # todo: move into tool

        from ph.emulation.machine import vmCurrentTask
        try: task = vmCurrentTask()
        except AttributeError: return
        task.checkAccessSystem(['zones:autoquest'], 'create')

        from stuphmud.server.zones.specials.autoquest import loadQuest

        VALID_QUEST_TYPES = ['$builtin.ProtectedSenario']

        assert value['quest-type'] in VALID_QUEST_TYPES
        return loadQuest(value, source = kwd.get('document', []) + [name])

    def value(self, name, value, **kwd):
        return value # identity

    # def synthetic(self, name, value, **kwd):
    #     return Synthetic(value)

    def url(self, name, value, **kwd):
        # :security: XXX :skip: unsafe load package
        raise NotImplementedError('Blocking unsafe operation')

        from phsite.network import url, importView, patterns, include_urlpatterns
        try:
            try: value = value['include']
            except KeyError:
                view = value['view'].split('.')
                viewModule = '.'.join(view[:-1])
                viewName = view[-1]

                view = importView(viewModule, viewName)
                try: view = view.view
                except AttributeError:
                    pass

                return url(value['pattern'], view)
            else:
                if isinstance(value, list):
                    return patterns('', value)

                from stuphmud.server.runtime import LookupObject
                include = LookupObject(value['package']).urlpatterns
                return include_urlpatterns(value['pattern'],
                                           include)

        except Exception as e:
            return e

        '''
        patterns:
          - (stuph$url):
            pattern: '^/html/format$'
            view: person.services.web.views.format_html
          - (stuph$url):
            include:
              pattern: '^game/'
              package: web.stuph.embedded.urls
          - (stuph$url):
            include:
            - (stuph$url):
              pattern: '^accounts/profile'
              view: web.stuph.accounts.profile
        '''

    def library(self, name, value, **kwd):
        # XXX :skip: This introduces a lowlevel object that has no memory management.
        raise NotImplementedError('Blocking unsafe operation')

        return LibraryCore(value.getSection('LibraryCore'),
                           LibraryCore.Node)

    class factory(Object, object):
        '''
        my/application:
            interfaces/interface::
                object($submapping):
                    classes:
                        object: my/components/object/kernel

                    document::
                        component($object): true


            documents/usage::
                structure('my/application/interface').object.component

                '''

        def __new__(self, name, value, **kwd):
            raise NotImplementedError

            core = runtime[runtime.Agent.System]
            if not core:
                raise NotImplementedError('%s not installed' % runtime.Agent.System)

            try: classes = value['classes']
            except NameError:
                classes = value
                document = None
            else:
                document = value['document']

            # Todo: merge parent-document 'value' structure into sub-document result object?

            f = object.__new__(self, core, name, classes, **kwd)
            f.__init__(core, name, classes, **kwd)

            if document is None:
                return f

            # Load sub-document.
            return f(document, **kwd) # kwd env?


        def __init__(self, core, name, classes, **kwd):
            self._classes = dict((n, self._lookupSubmapping(core, v))
                                 for (n, v) in classes.items())

        def _lookupSubmapping(self, core, n):
            s = n.split('/')

            for i in range(len(s), -1, -1):
                n = s[:i]

                try: n = core.lookup(*n) # XXX :skip: won't the Node.Tool just dereference its scope..?
                except KeyError:
                    continue

                if isinstance(n, core.Node.Tool):
                    n = n.scope

                    for o in s[i:]:
                        n = getattr(n, o) # todo: catch AttributeError and restart library search?


                    # An object was found, now decide what to do with it.

                    # if isinstance(n, Submapping):
                    #   return n


                    # SECURITY NOTE:  Because structures are intended to be loaded from the
                    # virtual machine, they are restricted to only loaded native tool library
                    # paths, which should only be producing object encapsulations.
                    #
                    # Additionally, native tool structural objects must be loaded from the
                    # tool's "_Factory" object, because structural item building methods do
                    # not take frame parameters.  Instead, the factory must detect the vm's
                    # task context instance if it needs to do frame-dependent actions.

                    return n._Factory


                if isinstance(n, core.Node.Module):
                    return self._activity(self, core, n, s)


        @property
        def _machine(self):
            from world import heartbeat as vm
            return vm
            return runtime[runtime.System.Engine]

        def __call__(self, source, **env):
            from stuphos.language.document.interface import load

            def coprocess():
                # XXX :skip: object-load/instantiation order is arbitrary!
                return load(source, self._classes, '', **env)

            try: task = self._machine.taskObject # contextObject.task
            except AttributeError:
                return coprocess()
            else:
                # a parallel document-loading routine object that submaps to library activities
                # and when loading, evaluates the instantiations asynchronously

                # XXX :skip: it's not asynchronous, it's callAsynchronous, or some bull*.
                task.callAsynchronous(coprocess)


        class _activity:
            def __init__(self, submapping, core, node, segments):
                self._submapping = submapping
                self._core = core
                self._node = node
                self._segments = segments

            def __call__(self, name, value, **kwd):
                from stuphos.kernel import LibraryNode, Script, findCurrentUser
                from world import heartbeat as vm

                with vm.Loading(Script, environ = kwd, user = findCurrentUser()) as task:
                    lib = LibraryNode._WithCore(self._core, self._node, self._segments)
                    instance = lib.instantiate(arguments = dict(value = value))

                    @task.onComplete
                    def buildComplete():
                        # Hack the stack to assert an exact task return value.
                        task.stack[:] = [instance]


                    # Wait for the virtual task to complete, and return the
                    # object result inserted from the activity instantation.
                    # Todo: errors.
                    return task.returnValue


    submapping = factory

    def structure(self, name, value, **kwd):
        '''
        my/application:
            interfaces/interface::
                object($structure):
                    classes:
                        # The path is to a Native Tool with a Factory definition.
                        '': my/components/object/kernel

                    document::
                        # object is defined in the native tool factory.
                        component($object): true


            documents/usage::
                structure('my/application/interface').object.component


        '''

        # This can wrap a factory.
        raise NotImplementedError('Blocking unsafe operation')

        core = runtime[runtime.Agent.System]
        if core is not None:
            classes = value['classes']
            document = value['document']
            default = value.get('default', '') # 'stuph'

            if isinstance(classes, self.factory):
                classes = classes._classes
            else:
                def factory(path):
                    node = core[path]
                    if isinstance(node, core.Node.Module):
                        # Todo: invoke method (like call) within activity.
                        # Because this means raising a Continuation that a
                        # would require a rewrite of the document loading
                        # code, we can do it synchronously, waiting on
                        # invocations in the outer-frame but this must be
                        # done from a non-heartbeat thread!  So, move the
                        # document load into a pooled thread executor.
                        pass
                    elif isinstance(node, core.Node.Structure):
                        pass

                    elif isinstance(node, core.Node.Tool):
                        # XXX :skip: check permission.
                        return node.scope._Factory

                classes = dict((type, factory(path))
                                for (type, path) in classes)

            # from mud.lang.structure import Factory
            # classes['stuph'] = Factory

            from stuphos.language.document.interface import load
            return load(document, classes, default)


    class _MethodInterface(Subroutine._Interface):
        # As a Subroutine.Interface, this object is callable by the emulation
        # runtime by dereferencing the procedure.  This object generates new
        # subroutine instances on property dereference for individual frames.

        def __init__(self, module, name, **env):
            self._module = module
            self._name = name
            self._environ = env

        @property
        def _programmer(self):
            # Used by call instruction.
            return self._environ.get('programmer')

        @property
        def _subroutine(self):
            # debugOn()
            proc = self._module.getSubroutine(self._name)
            proc._setEnvironment(self._module._getEnvironment())
            return proc
        _procedure = _subroutine


        # XXX Should these exist?
        def _getEnvironment(self, *args, **kwd):
            proc = self._module.getSubroutine(self._name)
            return proc._getEnvironment(*args, **kwd)

        def setEnvironment(self, *args, **kwd):
            proc = self._module.getSubroutine(self._name)
            return proc.setEnvironment(*args, **kwd)


        @property
        def _activation(self):
            from ph.emulation.operation.application import Invocation
            proc = self._module.getSubroutine(self._name)
            # debugOn()
            return Invocation(proc.main_program, proc.name, proc.start_position)

        def __repr__(self):
            return f'<method$interface {repr(self._name)}>'


    def method(self, name, value, **kwd):
        # Not really sure what the point of this is, to be part of
        # a synthetic (structural) class.  It could be wrapped by
        # something to mark it and get passed to the class constructor,
        # but, a method definition really has no use alone.
        module = buildMethodFromStructure(name, value, **kwd)

        # print('[structure.method]')
        # print(indent(module))

        from stuphos.kernel import Girl, FastLocals, GrammaticalError
        from stuphos.language.document.interface import getContextEnvironment

        try: module = Girl(Girl.Module, module, types = \
            getContextEnvironment('types', default = None))

        except GrammaticalError as e:
            raise ValueError(module) from e

        env = dict(container = containerAccessor(kwd['container']))

        try: task = vmCurrentTask()
        except AttributeError:
            task = None

        else:
            env = FastLocals(task.memory, env, module._variableNames,
                             module._variableIndices)
            # print(f'[method$locals] {id(env)}')

        # print(f'[method def] {", ".join(kwd.keys())}')

        try: path = getContextEnvironment('document')
        except KeyError: pass
        else:
            module._setAttribute('_interface_path', '%s/%s' % ('/'.join(path), name))

            if task:
                env['path$'] = task.sequence(path + [name])

        module.setEnvironment(env)

        # def getEnvironmentVariable(name):
        #     return getContextEnvironment('document')[name]

        # module.environ['getenv'] = _safe_native(getEnvironmentVariable)

        # module.setEnvironment(vmCurrentTask().environ)

        # Get the owner of the (agent system) interface, if it exists.
        # from stuphos.language.document.structural import Context
        # progr = Context['loader'].environ.get('programmer')

        progr = getContextEnvironment('programmer', default = None)

        return self._MethodInterface(module, name, programmer = progr)

    procedure = subroutine = method


    # Currently fails.  Also, returns a non-native-safe object.
    # from stuphos.language.shortate import shortate
    # shortate = staticmethod(shortate)


    def command(self, name, value, **kwd):
        # XXX :skip: Provide a memory-sensitive implementation of actionable.command[.verb]
        raise NotImplementedError('Blocking unsafe operation')

        # Requires implementation.
        from spatial.architecture import actionable

        if isinstance(value, str):
            value = dict(implementation = value)

        parse = value.get('parse')

        if isinstance(parse, str):
            parse = buildMethodFromStructure(parse, **kwd)

        impl = value['implementation']

        if isinstance(impl, str):
            impl = buildMethodFromStructure(impl, **kwd)

        # if isinstance(parse, subroutine):
        #   parse = compileSubroutine(parse)
        # if isinstance(impl, subroutine):
        #   impl = compileSubroutine(impl)


        ns = dict(__call__ = impl)
        if parse is not None:
            ns['parseArgs'] = parse


        verbClass = newClassType('implementation', (actionable.command.verb,), ns)
        ns = dict(name = name, implementation = verbClass)

        return newClassType('%sCommand' % name, (actionable.command,), ns)


identifier_pattern = re.compile('[a-zA-Z$_][a-zA-Z$_0-9]*').match
def isValidIdentifier(name):
    return identifier_pattern(name) is not None

def buildMethodFromStructure(name, value, **kwd):
    if not isValidIdentifier(name):
        # If it's coming from structure, validate.
        raise NameError(name)

    if isinstance(value, str):
        parameters = []
        code = value
    else:
        parameters = value['parameters']
        code = value['code']

    parameters = ', '.join(parameters)
    return buildMethod(name, parameters, code)

def buildMethod(name, parameters, code = 'pass'):
    # from ph.lang.layer import grammar
    # ast = grammar.Grammar.AST

    # suite = ast.Suite([])
    # module = ast.Suite([ast.FunctionDefinition(name, parameters, suite)])

    code = 'locals().update(keywords$())\n' + code

    return 'def %s(%s):\n%s\n' % (name, parameters, indent(code))


@staticmethod
def htmlScript(name, value, **kwd):
    # XXX :skip: value must be interned here, meaning a memory mapping or sequence,
    # otherwise, structural.Items won't be constrained by mental.objects code.
    raise NotImplementedError('Blocking unsafe operation')

    return HtmlView._ScriptElement(value)

#@staticmethod
def syntheticClass(self, name, value, **kwd):
    '''
    code.value.object.create()() <- code:
        object(class)::
            def call$():
                pass

    '''

    from stuphos.kernel import Girl, SyntheticClass

    # todo: $classInit$ trigger method for initializing
    # environment.  Also, pass kwd/document container
    # for connecting class impl to structure.

    # Todo: use SyntheticClass._BuildProgram/FromStructure
    # This however utilizes SyntheticProgram, which is different
    # from the source-composing strategy happening here.
    #
    # The FromStructure strategy is lower level, although doesn't
    # compose buildMethodFromStructure values and doesn't allow
    # environment.

    attributes = []

    if isinstance(value, str):
        init = value
    else:
        def checkMember(n, m):
            if isinstance(m, (dict, str)):
                return True

            # debugOn()

            attributes.append((n, m))
            return False

        methods = ((n, m) for (n, m) in value.items()
                   if checkMember(n, m))

        init = nls(buildMethodFromStructure(defn, m, **kwd) for
                   (defn, m) in methods)

    types = getContextEnvironment('types', default = None)

    synth = SyntheticClass \
            (name,
             Girl(Girl.Module, init, types = types),
             kwd.get('container'))

    # for (n, m) in attributes:
    #     # todo: is this right?  Shouldn't it be set in environ?
    #     synth._setAttribute(n, m)

    synth._subroutines = dict(attributes)

    # Note: classInit must be called by client.
    return synth


setattr(Factory, 'html.script', htmlScript)
setattr(Factory, 'class', syntheticClass)
Factory.object = Factory.formula = getattr(Factory, 'class')

from stuphos.kernel import GirlCore
class LibraryCore(GirlCore):
    def loadEntities(self, cfg, nodeClass):
        pass
    def saveNode(self, node, recurse = False):
        pass
    def destroyNode(self, node):
        pass


class SystemFactory(Submapping):
    def tool(self, path, sourceCode, **kwd):
        segments = path.split('/')
        assert isinstance(sourceCode, str)
        core = runtime[runtime.Agent.System]
        assert core is not None
        scope = Core.Python().module(path, sourceCode,
                                     container = kwd['container'])
        return core.addPythonTool(segments, scope) # Q: dictOf()?

    # def lock(self, name, value, **kwd):
    #   from threading import Lock
    #   return Lock()


class HTMLFactory(Submapping):
    # XXX __str__ continuation not supported.
    class _iframe(writeprotected):
        def __init__(self, name, value, **kwd):
            if isinstance(value, dict):
                src = value['source']

            elif isinstance(value, str):
                src = value
            else:
                raise ValueError(value)

            self._src = src

        def buildSource(self, src, attrs = ''):
            # todo: escapeXmlContent(src)
            return f'<iframe{attrs and " " or ""}{attrs} src="data:text/html,{src}"></iframe>'

        def build(self):
            src = self._src
            if isinstance(src, str):
                return self.buildSource(src)

            # elif callable(src):
            else:
                new = vmCurrentTask().swapFrameCall(resolve_procedure(src))

                @new.onComplete
                def run(frame, *error, **kwd):
                    stack = frame.task.stack
                    stack.push(self.buildSource(stack.pop()[0]))

                raise new.outer

        __str__ = __call__ = build


        # from stuphos.runtime.architecture.api import _safe_native
        # return _safe_native(build)


from stuphos.language.document.interface import getContextEnvironment
